@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  */
 
 #include <sys/types.h>
@@ -81,7 +81,7 @@ duprow()
 	if (p) {
 	    register struct ent *n;
 	    n = lookat(currow, curcol);
-	    (void) copyent(n, p, 1, 0);
+	    (void) copyent(n, p, 1, 0, 0, 0, maxrow, maxcol, 0);
 	}
     }
     for (curcol = c1; curcol <= c2; curcol++) {
@@ -109,7 +109,7 @@ dupcol()
 	if (p) {
 	    register struct ent *n;
 	    n = lookat(currow, curcol);
-	    copyent(n, p, 0, 1);
+	    copyent(n, p, 0, 1, 0, 0, maxrow, maxcol, 0);
 	}
     }
     for (currow = 0; currow <= maxrow; currow++) {
@@ -142,13 +142,12 @@ insertrow(int arg, int delta)
     if ((fr = find_frange(currow + delta, curcol))) {
 	move_area(currow + arg + delta, fr->or_left->col, currow + delta,
 		fr->or_left->col, fr->or_right->row, fr->or_right->col);
-	flush_saved();
 	if (!delta && fr->ir_left->row == currow + arg)
 	    fr->ir_left = lookat(fr->ir_left->row - arg, fr->ir_left->col);
 	if (delta && fr->ir_right->row == currow)
 	    fr->ir_right = lookat(fr->ir_right->row + arg, fr->ir_right->col);
 
-	for (i=0; i < 27; i++) {	/* update all marked cells */
+	for (i = 0; i < 37; i++) {	/* update all marked cells */
 	    if (savedrow[i] >= currow + delta &&
 		    savedcol[i] >= fr->or_left->col &&
 		    savedcol[i] <= fr->or_right->col)
@@ -189,7 +188,7 @@ insertrow(int arg, int delta)
 	}
 	tbl[r] = tmprow;		/* the last row was never used.... */
 
-	for (i=0; i < 27; i++) {	/* update all marked cells */
+	for (i = 0; i < 37; i++) {	/* update all marked cells */
 	    if (savedrow[i] >= currow + delta)
 		savedrow[i] += arg;
 	    if (savedstrow[i] >= currow + delta)
@@ -263,7 +262,7 @@ insertcol(int arg, int delta)
     }
 
     /* Update all marked cells. */
-    for (c=0; c < 27; c++) {
+    for (c=0; c < 37; c++) {
 	if (savedcol[c] >= curcol + delta)
 	    savedcol[c] += arg;
 	if (savedstcol[c] >= curcol + delta)
@@ -305,10 +304,12 @@ insertcol(int arg, int delta)
 void
 deleterow(register int arg)
 {
-    int rs = maxrow - currow + 1;
     int i;
-    char buf[50];
+    int rs = maxrow - currow + 1;
     struct frange *fr;
+    struct ent *p;
+    struct ent *obuf = NULL;
+    char buf[50];
 
     if ((fr = find_frange(currow, curcol)))
 	rs = fr->or_right->row - currow + 1;
@@ -326,23 +327,45 @@ deleterow(register int arg)
 	else {
 	    FullUpdate++;
 	    modflg++;
+	    if (dbidx < 0) dbidx++;
+	    delbuf[dbidx] = delbuf[DELBUFSIZE - 1];
+	    delbuf[DELBUFSIZE - 1] = NULL;
+	    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+		if (delbuf[i] == delbuf[dbidx]) {
+		    delbuf[dbidx] = NULL;
+		    break;
+		}
+	    }
 	    flush_saved();
+	    if (qbuf) {
+		if (dbidx < 0) dbidx++;
+		delbuf[dbidx] = delbuf[qbuf];
+		flush_saved();
+		obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+	    }
 	    sync_refs();
 	    erase_area(currow, fr->or_left->col, currow + arg - 1,
 		    fr->or_right->col);
 	    fix_ranges(currow, -1, currow + arg - 1, -1, -1, -1);
+	    for (i = 0; i < DELBUFSIZE; i++)
+		if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+		    delbuf[i] = delbuf[dbidx];
+	    qbuf = 0;
+	    for (i = DELBUFSIZE - 1; i > DELBUFSIZE - 9; i--)
+		delbuf[i] = delbuf[i-1];
+	    delbuf[DELBUFSIZE - 9] = delbuf[dbidx];
+	    for (p = delbuf[dbidx]; p; p = p->next)
+		p->flags &= ~may_sync;
 	    if (currow + arg > fr->or_right->row)
 		fr->or_right = lookat(currow - 1, fr->or_right->col);
-	    else {
+	    else
 		move_area(currow, fr->or_left->col, currow + arg,
 			fr->or_left->col, fr->or_right->row, fr->or_right->col);
-		flush_saved();
-	    }
 	    if (currow + arg > fr->ir_right->row && fr->ir_right->row >= currow)
 		fr->ir_right = lookat(currow - 1, fr->ir_right->col);
 
 	    /* Update all marked cells. */
-	    for (i=0; i < 27; i++) {
+	    for (i = 0; i < 37; i++) {
 		if (savedcol[i] >= fr->or_left->col &&
 			savedcol[i] <= fr->or_right->col) {
 		    if (savedrow[i] >= currow && savedrow[i] < currow + arg)
@@ -387,11 +410,35 @@ deleterow(register int arg)
 	if (any_locked_cells(currow, 0, currow + arg - 1, maxcol))
 	    error("Locked cells encountered. Nothing changed");
 	else {
+	    if (dbidx < 0) dbidx++;
+	    delbuf[dbidx] = delbuf[DELBUFSIZE - 1];
+	    delbuf[DELBUFSIZE - 1] = NULL;
+	    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+		if (delbuf[i] == delbuf[dbidx]) {
+		    delbuf[dbidx] = NULL;
+		    break;
+		}
+	    }
 	    flush_saved();
+	    if (qbuf) {
+		if (dbidx < 0) dbidx++;
+		delbuf[dbidx] = delbuf[qbuf];
+		flush_saved();
+		obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+	    }
 	    sync_refs();
 	    erase_area(currow, 0, currow + arg - 1, maxcol);
 	    fix_ranges(currow, -1, currow + arg - 1, -1, -1, -1);
 	    closerow(currow, arg);
+	    for (i = 0; i < DELBUFSIZE; i++)
+		if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+		    delbuf[i] = delbuf[dbidx];
+	    qbuf = 0;
+	    for (i = DELBUFSIZE - 1; i > DELBUFSIZE - 9; i--)
+		delbuf[i] = delbuf[i-1];
+	    delbuf[DELBUFSIZE - 9] = delbuf[dbidx];
+	    for (p = delbuf[dbidx]; p; p = p->next)
+		p->flags &= ~may_sync;
 	}
     }
 }
@@ -400,8 +447,10 @@ void
 yankrow(int arg)
 {
     int rs = maxrow - currow + 1;
+    int i, qtmp;
     char buf[50];
     struct frange *fr;
+    struct ent *obuf;
 
     if ((fr = find_frange(currow, curcol)))
 	rs = fr->or_right->row - currow + 1;
@@ -412,23 +461,46 @@ yankrow(int arg)
 	error(buf);
 	return;
     }
-    if (fr) {
-	sync_refs();
+    sync_refs();
+    if (dbidx < 0) dbidx++;
+    delbuf[dbidx] = delbuf[DELBUFSIZE - 10];
+    delbuf[DELBUFSIZE - 10] = NULL;
+    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+	if (delbuf[i] == delbuf[dbidx]) {
+	    delbuf[dbidx] = NULL;
+	    break;
+	}
+    }
+    flush_saved();
+    if (qbuf) {
+	if (dbidx < 0) dbidx++;
+	delbuf[dbidx] = delbuf[qbuf];
 	flush_saved();
+	obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+    }
+    qtmp = qbuf;
+    qbuf = 0;
+    if (fr) {
 	yank_area(currow, fr->or_left->col, currow + arg - 1,
 		fr->or_right->col);
     } else {
-	sync_refs();
-	flush_saved();
 	yank_area(currow, 0, currow + arg - 1, maxcol);
     }
+    qbuf = qtmp;
+    for (i = 0; i < DELBUFSIZE; i++)
+	if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+	    delbuf[i] = delbuf[dbidx];
+    qbuf = 0;
+    delbuf[DELBUFSIZE - 10] = delbuf[dbidx];
 }
 
 void
 yankcol(int arg)
 {
     int cs = maxcol - curcol + 1;
+    int i, qtmp;
     char buf[50];
+    struct ent *obuf;
 
     if (cs - arg < 0) {
     	cs = cs > 0 ? cs : 0;
@@ -438,8 +510,31 @@ yankcol(int arg)
 	return;
     }
     sync_refs();
+    if (dbidx < 0) dbidx++;
+    delbuf[dbidx] = delbuf[DELBUFSIZE - 10];
+    delbuf[DELBUFSIZE - 10] = NULL;
+    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+	if (delbuf[i] == delbuf[dbidx]) {
+	    delbuf[dbidx] = NULL;
+	    break;
+	}
+    }
     flush_saved();
+    if (qbuf) {
+	if (dbidx < 0) dbidx++;
+	delbuf[dbidx] = delbuf[qbuf];
+	flush_saved();
+	obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+    }
+    qtmp = qbuf;
+    qbuf = 0;
     yank_area(0, curcol, maxrow, curcol + arg - 1);
+    qbuf = qtmp;
+    for (i = 0; i < DELBUFSIZE; i++)
+	if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+	    delbuf[i] = delbuf[dbidx];
+    qbuf = 0;
+    delbuf[DELBUFSIZE - 10] = delbuf[dbidx];
 }
 
 void
@@ -507,7 +602,7 @@ yank_area(int sr, int sc, int er, int ec)
     curcol = sc;
 
     erase_area(sr, sc, er, ec);
-    pullcells('m');
+    pullcells('p');
 
     currow = r;
     curcol = c;
@@ -541,13 +636,9 @@ move_area(int dr, int dc, int sr, int sc, int er, int ec)
     curcol = sc;
 
     /* First we erase the source range, which puts the cells on the delete
-     * buffer stack.  The pullcells(0) then makes a copy in the original
-     * location, which we then erase, putting the copy on the delete buffer
-     * stack, too.
+     * buffer stack.
      */
     erase_area(sr, sc, er, ec);
-    pullcells(0);
-    erase_area(sr, sc, er, ec);
 
     currow = r;
     curcol = c;
@@ -557,19 +648,17 @@ move_area(int dr, int dc, int sr, int sc, int er, int ec)
     /* Now we erase the destination range, which adds it to the delete buffer
      * stack, but then we flush it off.  We then move the original source
      * range from the stack to the destination range, adjusting the addresses
-     * as we go, which leaves only the copy on the stack for future pulls.
+     * as we go, leaving the stack in its original state.
      */
     erase_area(dr, dc, er + deltar, ec + deltac);
     flush_saved();
-    for (p = delbuf[dbidx - 1]; p; p = p->next) {
+    for (p = delbuf[dbidx]; p; p = p->next) {
 	pp = ATBL(tbl, p->row + deltar, p->col + deltac);
 	*pp = p;
 	p->row += deltar;
 	p->col += deltac;
 	p->flags &= ~is_deleted;
     }
-
-    delbuf[dbidx - 1] = delbuf[dbidx];
     delbuf[dbidx--] = NULL;
 }
 
@@ -616,16 +705,23 @@ valueize_area(int sr, int sc, int er, int ec)
 void
 pullcells(int to_insert)
 {
+    struct ent *obuf;
     struct ent *p, *n;
     struct ent **pp;
     int deltar, deltac;
     int minrow, mincol;
     int mxrow, mxcol;
     int numrows, numcols;
+    int i;
     struct frange *fr;
+    
+    if (qbuf && delbuf[qbuf])
+	delbuf[++dbidx] = delbuf[qbuf];
+    obuf = delbuf[dbidx];	/* orig. contents of the del. buffer */
 
-    if (dbidx < 0) {
+    if ((qbuf && !delbuf[qbuf]) || dbidx < 0) {
 	error("No data to pull");
+	qbuf = 0;
 	return;
     }
 
@@ -643,6 +739,7 @@ pullcells(int to_insert)
 	    mincol = p->col;
 	if (p->col > mxcol)
 	    mxcol = p->col;
+	p->flags |= may_sync;
     }
 
     numrows = mxrow - minrow + 1;
@@ -663,14 +760,27 @@ pullcells(int to_insert)
     } else if (to_insert == 'x') {	/* Do an exchange. */
 	struct ent *tmp;
 
-	/* Save the original contents of the destination range in preparation
-	 * for the exchange.
+	/* Save the original contents of the destination range on the
+	 * delete buffer stack in preparation for the exchange, then swap
+	 * the top two pointers on the stack, so that the original cells
+	 * to be pulled are still on top.
 	 */
 	erase_area(minrow + deltar, mincol + deltac, mxrow + deltar,
 		mxcol + deltac);
 	tmp = delbuf[dbidx];
 	delbuf[dbidx] = delbuf[dbidx - 1];
 	delbuf[dbidx - 1] = tmp;
+    } else if (to_insert == 'p') {
+	erase_area(minrow + deltar, mincol + deltac, mxrow + deltar,
+		mxcol + deltac);
+	sync_refs();
+	flush_saved();
+    } else if (to_insert == 't') {
+	erase_area(minrow + deltar, mincol + deltac,
+		minrow + deltar + mxcol - mincol,
+		mincol + deltac + mxrow - minrow);
+	sync_refs();
+	flush_saved();
     }
 
     FullUpdate++;
@@ -680,31 +790,33 @@ pullcells(int to_insert)
      * range.
      */
     for (p = delbuf[dbidx]; p; p = p->next) {
-	if (to_insert == 't')
-	    /* Transpose rows and columns while pulling. */
+	if (to_insert == 't')	/* Transpose rows and columns while pulling. */
 	    n = lookat(minrow + deltar + p->col - mincol,
 		    mincol + deltac + p->row - minrow);
 	else
 	    n = lookat(p->row + deltar, p->col + deltac);
-	(void) clearent(n);
-	copyent(n, p, 0, 0);
-	n->flags = p->flags & ~is_deleted;
+	copyent(n, p, deltar, deltac, minrow, mincol, mxrow, mxcol, to_insert);
     }
 
     /* Now exchange them so that the original cells from the delete buffer
      * are in the destination range instead of the copies.  When doing a
      * "pull exchange" ("px" or "pullxchg"), exchange the original contents
      * of the destination range with the contents of the delete buffer
-     * instead.  Don't do this if called from move_area() or when transposing.
+     * instead.  Don't do this if transposing or merging.
      */
-    if (to_insert && to_insert != 't') {
+    if (to_insert != 't' && to_insert != 'm') {
 	if (to_insert == 'x') {
 	    struct ent *tmp = delbuf[dbidx];
 	    delbuf[dbidx] = delbuf[dbidx - 1];
 	    delbuf[dbidx - 1] = tmp;
 	} else
-	    erase_area(minrow + deltar, mincol + deltac, mxrow + deltar,
-		    mxcol + deltac);
+	    for (p = delbuf[dbidx++]; p; p = p->next) {
+		pp = ATBL(tbl, p->row + deltar, p->col + deltac);
+		if (*pp && !((*pp)->flags & is_locked)) {
+		    free_ent(*pp);
+		    *pp = NULL;
+		}
+	    }
 	for (p = delbuf[dbidx - 1]; p; p = p->next) {
 	    pp = ATBL(tbl, p->row + deltar, p->col + deltac);
 	    *pp = p;
@@ -724,7 +836,17 @@ pullcells(int to_insert)
 	    p->row -= deltar;
 	    p->col -= deltac;
 	}
-    }
+    } else
+	sync_refs();
+    
+    /* Now make sure all references to the pulled cells in all named buffers
+     * point to the new set of cells in the delete buffer.
+     */
+    for (i = 0; i < DELBUFSIZE; i++)
+	if (delbuf[i] == obuf) delbuf[i] = delbuf[dbidx];
+    if (qbuf && delbuf[qbuf])
+	delbuf[dbidx--] = NULL;
+    qbuf = 0;
 }
 
 void
@@ -783,7 +905,11 @@ get_rcqual(int ch)
 {
     int c;
 
-    error("%sow/column:  r: row  c: column%s",
+    if (ch == 'p')
+	error("Pull:  r: row  c: column  p: paste  m: merge  x: exchange"
+		"  t: transpose");
+    else
+	error("%sow/column:  r: row  c: column",
 
 #ifdef KEY_IC
 	    (ch == KEY_IC)	? "Insert r" :
@@ -793,12 +919,9 @@ get_rcqual(int ch)
 	    (ch == 'a')		? "Append r" :
 	    (ch == 'd')		? "Delete r" :
 	    (ch == 'y')		? "Yank r" :
-	    (ch == 'p')		? "Pull r" :
 	    (ch == 'v')		? "Values r" :
 	    (ch == 'Z')		? "Zap r" :
-	    (ch == 's')		? "Show r" : "R",
-
-	    (ch == 'p') ? "  m: merge  x: exchange  t: transpose" : "");
+	    (ch == 's')		? "Show r" : "R");
 
     (void) refresh();
 
@@ -808,6 +931,8 @@ get_rcqual(int ch)
 
 	case 'c':	return ('c');
 
+	case 'p':	return ((ch == 'p') ? 'p' : 0);
+
 	case 'm':	return ((ch == 'p') ? 'm' : 0);
 
 	case 'x':	return ((ch == 'p') ? 'x' : 0);
@@ -826,8 +951,8 @@ get_rcqual(int ch)
 			    return (0);
 
 	case 'y':	if (ch == 'y') {
-			    flush_saved();
-			    yank_area(currow, curcol, currow, curcol);
+			    yankr(lookat(currow, curcol),
+				    lookat(currow, curcol));
 			    return (ESC);
 			} else
 			    return (0);
@@ -930,7 +1055,7 @@ closerow(int rs, int numrow)
     }
 
     /* Update all marked cells. */
-    for (i=0; i < 27; i++) {
+    for (i = 0; i < 37; i++) {
 	if (savedrow[i] >= rs && savedrow[i] < rs + numrow)
 	    savedrow[i] = savedcol[i] = -1;
 	if (savedrow[i] >= rs + numrow)
@@ -984,9 +1109,10 @@ void
 closecol(int arg)
 {
     int r, c, i;
-    register struct ent **pp;
-    register struct ent *q;
     int cs = maxcol - curcol + 1;
+    struct ent **pp;
+    struct ent *p;
+    struct ent *obuf = NULL;
     char buf[50];
 
     if (cs - arg < 0) {
@@ -1000,17 +1126,41 @@ closecol(int arg)
 	error("Locked cells encountered. Nothing changed");
 	return;
     }
+    if (dbidx < 0) dbidx++;
+    delbuf[dbidx] = delbuf[DELBUFSIZE - 1];
+    delbuf[DELBUFSIZE - 1] = NULL;
+    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+	if (delbuf[i] == delbuf[dbidx]) {
+	    delbuf[dbidx] = NULL;
+	    break;
+	}
+    }
     flush_saved();
+    if (qbuf) {
+	if (dbidx < 0) dbidx++;
+	delbuf[dbidx] = delbuf[qbuf];
+	flush_saved();
+	obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+    }
     sync_refs();
     erase_area(0, curcol, maxrow, curcol + arg - 1);
     fix_ranges(-1, curcol, -1, curcol + arg - 1, -1, -1);
+    for (i = 0; i < DELBUFSIZE; i++)
+	if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+	    delbuf[i] = delbuf[dbidx];
+    qbuf = 0;
+    for (i = DELBUFSIZE - 1; i > DELBUFSIZE - 9; i--)
+	delbuf[i] = delbuf[i-1];
+    delbuf[DELBUFSIZE - 9] = delbuf[dbidx];
+    for (p = delbuf[dbidx]; p; p = p->next)
+	p->flags &= ~may_sync;
 
     /* clear then copy the block left */
     cs = maxcols - arg - 1;
     for (r=0; r<=maxrow; r++) {
 	for (c = curcol; c - curcol < arg; c++)
-	    if ((q = *ATBL(tbl, r, c)))
-		free_ent(q);
+	    if ((p = *ATBL(tbl, r, c)))
+		free_ent(p);
 
 	pp = ATBL(tbl, r, curcol);
 	for (c=curcol; c <= cs; c++, pp++) {
@@ -1039,7 +1189,7 @@ closecol(int arg)
     }
 
     /* Update all marked cells. */
-    for (i=0; i < 27; i++) {
+    for (i = 0; i < 37; i++) {
 	if (savedcol[i] >= curcol && savedcol[i] < curcol + arg)
 	    savedrow[i] = savedcol[i] = -1;
 	if (savedcol[i] >= curcol + arg)
@@ -1094,11 +1244,8 @@ doend(int rowinc, int colinc)
     register struct ent *p;
     int r, c;
 
-    /* Remember the current position */
-    savedrow[0] = currow;
-    savedcol[0] = curcol;
-    savedstrow[0] = strow;
-    savedstcol[0] = stcol;
+    if (!loading)
+	remember(0);
 
     if (VALID_CELL(p, currow, curcol)) {
 	r = currow + rowinc;
@@ -1136,6 +1283,8 @@ doend(int rowinc, int colinc)
         }
 	rowsinrange = 1;
 	colsinrange = fwidth[curcol];
+	if (!loading)
+	    remember(1);
 
 	error ("");	/* clear line */
 	return;
@@ -1181,9 +1330,14 @@ doformat(int c1, int c2, int w, int p, int r)
 
     if (c1 >= maxcols && !growtbl(GROWCOL, 0, c1)) c1 = maxcols-1 ;
     if (c2 >= maxcols && !growtbl(GROWCOL, 0, c2)) c2 = maxcols-1 ;
+    
+    if (w == 0) {
+	error("Width too small - setting to 1");
+	w = 1;
+    }
 
     if (usecurses && w > COLS - rescol - 2) {
-	error("Format too large - Maximum = %d", COLS - rescol - 2);
+	error("Width too large - Maximum = %d", COLS - rescol - 2);
 	w = COLS - rescol - 2;
     }
 
@@ -1208,6 +1362,120 @@ doformat(int c1, int c2, int w, int p, int r)
     modflg++;
 }
 
+void
+formatcol(arg)
+{
+    int c;
+
+    error("Current format is %d %d %d",
+	    fwidth[curcol], precision[curcol],
+	    realfmt[curcol]);
+    update(0);
+    c = nmgetch();
+    while (c >= 0 && c != ctl('m') && c != 'q' && c != ESC &&
+	    c != ctl('g') && linelim < 0) {
+	if (c >= '0' && c <= '9') {
+	    int i;
+	    for (i = curcol; i < curcol + arg; i++)
+		realfmt[i] = c - '0';
+	} else
+	    switch (c) {
+		case KEY_LEFT:
+		case '<':
+		case 'h':
+		    {
+			int i;
+			for (i = curcol; i < curcol + arg; i++) {
+			    fwidth[i]--;
+			    if (fwidth[i] < 1)
+				fwidth[i] = 1;
+			}
+		    }
+		    modflg++;
+		    break;
+		case KEY_RIGHT:
+		case '>':
+		case 'l':
+		    {
+			int i;
+			for (i = curcol; i < curcol + arg; i++) {
+			    fwidth[i]++;
+			    if (fwidth[i] > COLS - rescol - 2)
+				fwidth[i] = COLS - rescol - 2;
+			}
+		    }
+		    modflg++;
+		    break;
+		case KEY_DOWN:
+		case '-':
+		case 'j':
+		    {
+			int i;
+			for (i = curcol; i < curcol + arg; i++) {
+			    precision[i]--;
+			    if (precision[i] < 0)
+				precision[i] = 0;
+			}
+		    }
+		    modflg++;
+		    break;
+		case KEY_UP:
+		case '+':
+		case 'k':
+		    {
+			int i;
+			for (i = curcol; i < curcol + arg; i++)
+			    precision[i]++;
+		    }
+		    modflg++;
+		    break;
+		case ' ':
+		    if (arg == 1)
+			(void) sprintf(line,
+				"format [for column] %s ",
+				coltoa(curcol));
+		    else {
+			(void) sprintf(line,
+				"format [for columns] %s:",
+				coltoa(curcol));
+			(void) sprintf(line+strlen(line), "%s ",
+				coltoa(curcol+arg-1));
+		    }
+		    linelim = strlen(line);
+		    insert_mode();
+		    error("Current format is %d %d %d",
+			    fwidth[curcol], precision[curcol],
+			    realfmt[curcol]);
+		    continue;
+		case '=':
+		    error("Define format type (0-9):");
+		    update(0);
+		    if ((c = nmgetch()) >= '0' && c <= '9') {
+			(void) sprintf(line,
+				"format %c = \"", c);
+			linelim = strlen(line);
+			insert_mode();
+			error("");
+		    } else {
+			error("Invalid format type");
+			c = -1;
+		    }
+		    continue;
+		default:
+		    break;
+	    }
+	error("Current format is %d %d %d",
+		fwidth[curcol], precision[curcol],
+		realfmt[curcol]);
+	FullUpdate++;
+	update(1);
+	if (linelim < 0)
+	    c = nmgetch();
+    }
+    if (c >= 0)
+	error("");
+}
+
 void
 print_options(FILE *f)
 {
@@ -1222,7 +1490,6 @@ print_options(FILE *f)
 	!numeric &&
 	prescale == 1.0 &&
 	!extfunc &&
-	showcell &&
 	showtop &&
 	tbl_style == 0 &&
 	craction == 0 &&
@@ -1256,8 +1523,6 @@ print_options(FILE *f)
 	(void) fprintf(f, " prescale");
     if (extfunc)
 	(void) fprintf(f, " extfun");
-    if (!showcell)
-	(void) fprintf(f, " !cellcur");
     if (!showtop)
 	(void) fprintf(f, " !toprow");
     if (tbl_style)
@@ -1749,7 +2014,8 @@ unspecial(FILE *f, char *str, int delim)
 }
 
 struct enode *
-copye(register struct enode *e, int Rdelta, int Cdelta)
+copye(register struct enode *e, int Rdelta, int Cdelta, int r1, int c1,
+	int r2, int c2, int transpose)
 {
     register struct enode *ret;
 
@@ -1763,16 +2029,32 @@ copye(register struct enode *e, int Rdelta, int Cdelta)
 	} else
 	    ret = (struct enode *) scxmalloc((unsigned) sizeof (struct enode));
 	ret->op = e->op;
-	newrow=e->e.r.left.vf & FIX_ROW ? e->e.r.left.vp->row :
-					  e->e.r.left.vp->row+Rdelta;
-	newcol=e->e.r.left.vf & FIX_COL ? e->e.r.left.vp->col :
-					  e->e.r.left.vp->col+Cdelta;
+	newrow = e->e.r.left.vf & FIX_ROW ||
+		 e->e.r.left.vp->row < r1 || e->e.r.left.vp->row > r2 ||
+		 e->e.r.left.vp->col < c1 || e->e.r.left.vp->col > c2 ?
+		 e->e.r.left.vp->row :
+		 transpose ? r1 + Rdelta + e->e.r.left.vp->col - c1 :
+		 e->e.r.left.vp->row + Rdelta;
+	newcol = e->e.r.left.vf & FIX_COL ||
+		 e->e.r.left.vp->row < r1 || e->e.r.left.vp->row > r2 ||
+		 e->e.r.left.vp->col < c1 || e->e.r.left.vp->col > c2 ?
+		 e->e.r.left.vp->col :
+		 transpose ? c1 + Cdelta + e->e.r.left.vp->row - r1 :
+		 e->e.r.left.vp->col + Cdelta;
 	ret->e.r.left.vp = lookat(newrow, newcol);
 	ret->e.r.left.vf = e->e.r.left.vf;
-	newrow=e->e.r.right.vf & FIX_ROW ? e->e.r.right.vp->row :
-					   e->e.r.right.vp->row+Rdelta;
-	newcol=e->e.r.right.vf & FIX_COL ? e->e.r.right.vp->col :
-					   e->e.r.right.vp->col+Cdelta;
+	newrow = e->e.r.right.vf & FIX_ROW ||
+		 e->e.r.right.vp->row < r1 || e->e.r.right.vp->row > r2 ||
+		 e->e.r.right.vp->col < c1 || e->e.r.right.vp->col > c2 ?
+		 e->e.r.right.vp->row :
+		 transpose ? r1 + Rdelta + e->e.r.right.vp->col - c1 :
+		 e->e.r.right.vp->row + Rdelta;
+	newcol = e->e.r.right.vf & FIX_COL ||
+		 e->e.r.right.vp->row < r1 || e->e.r.right.vp->row > r2 ||
+		 e->e.r.right.vp->col < c1 || e->e.r.right.vp->col > c2 ?
+		 e->e.r.right.vp->col :
+		 transpose ? c1 + Cdelta + e->e.r.right.vp->row - r1 :
+		 e->e.r.right.vp->col + Cdelta;
 	ret->e.r.right.vp = lookat(newrow, newcol);
 	ret->e.r.right.vf = e->e.r.right.vf;
     } else {
@@ -1786,10 +2068,18 @@ copye(register struct enode *e, int Rdelta, int Cdelta)
 	    case 'v':
 		{
 		    int newrow, newcol;
-		    newrow=e->e.v.vf & FIX_ROW ? e->e.v.vp->row :
-						 e->e.v.vp->row+Rdelta;
-		    newcol=e->e.v.vf & FIX_COL ? e->e.v.vp->col :
-						 e->e.v.vp->col+Cdelta;
+		    newrow = e->e.v.vf & FIX_ROW ||
+			     e->e.v.vp->row < r1 || e->e.v.vp->row > r2 ||
+			     e->e.v.vp->col < c1 || e->e.v.vp->col > c2 ?
+			     e->e.v.vp->row :
+			     transpose ?  r1 + Rdelta + e->e.v.vp->col - c1 :
+			     e->e.v.vp->row + Rdelta;
+		    newcol = e->e.v.vf & FIX_COL ||
+			     e->e.v.vp->row < r1 || e->e.v.vp->row > r2 ||
+			     e->e.v.vp->col < c1 || e->e.v.vp->col > c2 ?
+			     e->e.v.vp->col :
+			     transpose ?  c1 + Cdelta + e->e.v.vp->row - r1 :
+			     e->e.v.vp->col + Cdelta;
 		    ret->e.v.vp = lookat(newrow, newcol);
 		    ret->e.v.vf = e->e.v.vf;
 		    break;
@@ -1798,16 +2088,21 @@ copye(register struct enode *e, int Rdelta, int Cdelta)
 		ret->e.k = e->e.k;
 		break;
 	    case 'f':
-		ret->e.o.right = copye(e->e.o.right,0,0);
+		ret->e.o.right = copye(e->e.o.right, 0, 0, r1, c1, r2, c2,
+			transpose);
 		ret->e.o.left = (struct enode *)0;
  		break;
 	    case '$':
+	    case EXT:
 		ret->e.s = scxmalloc((unsigned) strlen(e->e.s)+1);
 		(void) strcpy(ret->e.s, e->e.s);
-		break;
+		if (e->op == '$')	/* Drop through if ret->op is EXT */
+		    break;
 	    default:
-		ret->e.o.right = copye(e->e.o.right,Rdelta,Cdelta);
-		ret->e.o.left = copye(e->e.o.left,Rdelta,Cdelta);
+		ret->e.o.right = copye(e->e.o.right, Rdelta, Cdelta,
+			r1, c1, r2, c2, transpose);
+		ret->e.o.left = copye(e->e.o.left, Rdelta, Cdelta,
+			r1, c1, r2, c2, transpose);
 		break;
 	}
     }
@@ -1843,13 +2138,12 @@ syncref(register struct enode *e)
     } else {
 	switch (e->op) {
 	    case 'v':
-		if (!(e->e.v.vp->flags & is_cleared))
-		    e->e.v.vp = lookat(e->e.v.vp->row, e->e.v.vp->col);
-		else {
+		if (e->e.v.vp->flags & is_cleared) {
 		    e->op = ERR_;
 		    e->e.o.left = NULL;
 		    e->e.o.right = NULL;
-		}
+		} else if (e->e.v.vp->flags & may_sync)
+		    e->e.v.vp = lookat(e->e.v.vp->row, e->e.v.vp->col);
 		break;
 	    case 'k':
 		break;
@@ -2067,26 +2361,49 @@ closefile(FILE *f, int pid, int rfd)
     }
 }
 
+/* Copy a cell (struct ent).  "special" indicates special treatment when
+ * merging two cells for the "pm" command or for adjusting cell references
+ * when transposing with the "pt" command.  r1, c1, r2, and c2 define the
+ * range in which the dr and dc values should be used.
+ */
 void
-copyent(register struct ent *n, register struct ent *p, int dr, int dc)
+copyent(register struct ent *n, register struct ent *p, int dr, int dc,
+	int r1, int c1, int r2, int c2, int special)
 {
     if (!n || !p) {
 	error("internal error");
 	return;
     }
-    n->v = p->v;
-    n->flags = p->flags;
-    n->expr = copye(p->expr, dr, dc);
-    n->label = (char *)0;
+    if (special != 'm' || p->flags & is_valid) {
+	n->v = p->v;
+	n->flags |= p->flags & is_valid;
+    }
+    if (special != 'm' || p->expr) {
+	n->expr = copye(p->expr, dr, dc, r1, c1, r2, c2, special == 't');
+	if (p->flags & is_strexpr)
+	    n->flags |= is_strexpr;
+	else
+	    n->flags &= ~is_strexpr;
+    }
     if (p->label) {
+	if (n->label)
+	    scxfree(n->label);
 	n->label = scxmalloc((unsigned) (strlen(p->label) + 1));
 	(void) strcpy(n->label, p->label);
+	n->flags &= ~is_leftflush;
+	n->flags |= ((p->flags & is_label) | (p->flags & is_leftflush));
+    } else if (special != 'm') {
+	n->label = NULL;
+	n->flags &= ~(is_label | is_leftflush);
     }
-    n->format = 0;
     if (p->format) {
+	if (n->format)
+	    scxfree(n->format);
         n->format = scxmalloc((unsigned) (strlen(p->format) + 1));
 	(void) strcpy(n->format, p->format);
-    }
+    } else if (!(special == 'm'))
+	n->format = NULL;
+    n->flags |= is_changed;
 }
 
 void
@@ -2102,12 +2419,12 @@ write_fd(register FILE *f, int r0, int c0, int rn, int cn)
     for (c = 0; c < COLFORMATS; c++)
 	if (colformat[c])
 	    (void) fprintf (f, "format %d = \"%s\"\n", c, colformat[c]);
-    for (c = c0; c < cn; c++)
+    for (c = c0; c <= cn; c++)
 	if (fwidth[c] != DEFWIDTH || precision[c] != DEFPREC ||
 		realfmt[c] != DEFREFMT)
 	    (void) fprintf (f, "format %s %d %d %d\n", coltoa(c), fwidth[c],
 		    precision[c], realfmt[c]);
-    for (c = c0; c < cn; c++)
+    for (c = c0; c <= cn; c++)
         if (col_hidden[c])
             (void) fprintf(f, "hide %s\n", coltoa(c));
     for (r = r0; r <= rn; r++)
@@ -2143,10 +2460,6 @@ write_fd(register FILE *f, int r0, int c0, int rn, int cn)
 		}
 	    }
     }
-    /* xxx
-    if (rndtoeven)
-	fprintf(f, "set rndtoeven\n");
-    */
     /*
      * Don't try to combine these into a single fprintf().  v_name() has
      * a single buffer that is overwritten on each call, so the first part
@@ -2344,6 +2657,7 @@ readfile(char *fname, int eraseflg)
 
     if (eraseflg) erasedb();
 
+    remember(0);
     loading++;
     savefd = macrofd;
     macrofd = rfd;
@@ -2358,6 +2672,7 @@ readfile(char *fname, int eraseflg)
     }
     macrofd = savefd;
     --loading;
+    remember(1);
     closefile(f, pid, rfd);
 #ifndef MSDOS
     if (f == stdin) {
@@ -2370,9 +2685,10 @@ readfile(char *fname, int eraseflg)
 	(void) strcpy(curfile, save);
 	modflg = 0;
 	cellassign = 0;
-	if (autorun) readfile(autorun, 0);
+	if (autorun && !skipautorun) readfile(autorun, 0);
 	EvalAll();
     }
+    skipautorun = 0;
     autolabel = tempautolabel;
 }
 
@@ -2400,8 +2716,6 @@ erasedb()
 		*pp = (struct ent *)0;
 	    }
     }
-    for (dbidx = 3; dbidx >= 0; )
-	delbuf[dbidx--] = NULL;
 
     maxrow = 0;
     maxcol = 0;
@@ -2415,6 +2729,7 @@ erasedb()
     tbl_style = 0;
     craction = 0;
     rowlimit = collimit = -1;
+    qbuf = 0;
 
     autocalc=showcell=showtop=1;
     autoinsert=optimize=numeric=extfunc=color=colorneg=colorerr=cslop=0;
@@ -2422,15 +2737,9 @@ erasedb()
     if (usecurses && has_colors())
 	color_set(0, NULL);
     /* unset all marks */
-    for (c = 1; c < 27; c++)
+    for (c = 1; c < 37; c++)
 	savedrow[c] = savedcol[c] = savedstrow[c] = savedstcol[c] = -1;
 
-    /* set location to return to with `` or '' */
-    savedrow[0] = currow;
-    savedcol[0] = curcol;
-    savedstrow[0] = strow;
-    savedstcol[0] = stcol;
-
     if (mdir) {
 	scxfree(mdir);
 	mdir = NULL;
@@ -2553,7 +2862,7 @@ markcell()
 	return;
     }
     if ((c -= ('a' - 1)) < 1 || c > 26) {
-	error("Invalid character for mark (must be a-z)");
+	error("Invalid mark (must be a-z)");
 	return;
     }
     error("");
@@ -2568,24 +2877,19 @@ dotick(int tick)
 {
     int c;
 
-    /* Remember the current position */
-    int tmprow = currow;
-    int tmpcol = curcol;
-    int tmpstrow = strow;
-    int tmpstcol = stcol;
+    remember(0);
 
     error("Go to marked cell:");
     if ((c = nmgetch()) == ESC || c == ctl('g')) {
 	error("");
 	return;
     }
-    if ((c -= ('a' - 1)) < 1 || c > 26) {
-	if (tick == '`' || tick == '\'')
-	    c = 0;
-	else {
-	    error("Invalid character for mark (must be a-z, ` or \')");
-	    return;
-	}
+    if (c == '`' || c == '\'')
+	c = 0;
+    else if (!(((c -= ('a' - 1)) > 0 && c < 27) ||
+	    ((c += ('a' - '0' + 26)) > 26 && c < 37))) {
+	error("Invalid mark (must be a-z, 0-9, ` or \')");
+	return;
     }
     if (savedrow[c] == -1) {
 	error("Mark not set");
@@ -2602,10 +2906,7 @@ dotick(int tick)
 	gs.stflag = 1;
     } else
 	gs.stflag = 0;
-    savedrow[0] = tmprow;
-    savedcol[0] = tmpcol;
-    savedstrow[0] = tmpstrow;
-    savedstcol[0] = tmpstcol;
+    remember(1);
 
     FullUpdate++;
 }
