@@ -2,13 +2,13 @@
  * Help functions for sc 
  * R. Bond, 1988
  * J. Buhrt 1990
- * $Revision: 7.13 $
+ * $Revision: 7.15 $
  */
 
 #ifdef QREF
 #include <stdio.h>
 char	*header = " Quick Reference";
-char	*revision = "$Revision: 7.13 $";
+char	*revision = "$Revision: 7.15 $";
 #else
 #include <curses.h>
 #include "sc.h"
@@ -149,7 +149,7 @@ char *cell[] = {
 #endif
 "     =    Enter a numeric constant or expression.",
 "     <    Enter a left justified string or string expression.",
-"     \"    Enter a centered label.",
+"     \\    Enter a centered label.",
 "     >    Enter a right justified string or string expression.",
 "     e    Edit the current cell's numeric value.",
 "     E    Edit the current cell's string part.",
