@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  */
 
 #include <sys/types.h>
@@ -188,15 +188,18 @@ doindex(int minr, int minc, int maxr, int maxc, struct enode *val)
     int r, c;
     register struct ent *p;
 
-    if (minr == maxr) {			/* look along the row */
+    if (val->op == ',') {		/* index by both row and column */
+	r = minr + (int) eval(val->e.o.left) - 1;
+	c = minc + (int) eval(val->e.o.right) - 1;
+    } else if (minr == maxr) {		/* look along the row */
 	r = minr;
 	c = minc + (int) eval(val) - 1;
     } else if (minc == maxc) {		/* look down the column */
 	r = minr + (int) eval(val) - 1;
 	c = minc;
     } else {
-	r = minr + (int) eval(val->e.o.left) - 1;
-	c = minc + (int) eval(val->e.o.right) - 1;
+	error("Improper indexing operation");
+	return (double) 0;
     }
 
     if (c <= maxc && c >=minc && r <= maxr && r >=minr &&
@@ -825,7 +828,7 @@ eval(register struct enode *e)
 	    }
 	case FV:
 	case PV:
-	case PMT:	return (finfunc(e->op,eval(e->e.o.left),
+	case PMT:	return (finfunc(e->op, eval(e->e.o.left),
 				    eval(e->e.o.right->e.o.left),
 				    eval(e->e.o.right->e.o.right)));
 	case HOUR:	return (dotime(HOUR, eval(e->e.o.right)));
@@ -973,17 +976,20 @@ dofmt(char *fmtstr, double v)
 /*
  * Given a command name and a value, run the command with the given value and
  * read and return its first output line (only) as an allocated string, always
- * a copy of prevstr, which is set appropriately first unless external
+ * a copy of se->e.o.s, which is set appropriately first unless external
  * functions are disabled, in which case the previous value is used.  The
- * handling of prevstr and freeing of command is tricky.  Returning an
+ * handling of se->e.o.s and freeing of command is tricky.  Returning an
  * allocated string in all cases, even if null, insures cell expressions are
  * written to files, etc.
  */
 
 #if defined(VMS) || defined(MSDOS)
 char *
-doext(char *command, double value)
+doext(struct enode *se)
 {
+    char *command = seval(se->e.o.left);
+    double value = eval(se->e.o.right);
+
     error("Warning: External functions unavailable on VMS");
     cellerror = CELLERROR;	/* not sure if this should be a cellerror */
     if (command)
@@ -994,16 +1000,17 @@ doext(char *command, double value)
 #else /* VMS */
 
 char *
-doext(char *command, double value)
+doext(struct enode *se)
 {
-    static char *prevstr = (char *)0;	/* previous result */
-    static unsigned	prevlen = 0;
     char buff[FBUFLEN];		/* command line/return, not permanently alloc */
+    char *command;
+    double value;
 
+    command = seval(se->e.o.left);
+    value = eval(se->e.o.right);
     if (!extfunc) {
 	error("Warning: external functions disabled; using %s value",
-		((prevstr == NULL) || (*prevstr == '\0')) ?
-			"null" : "previous");
+		(se->e.o.s && *se->e.o.s) ? "previous" : "null");
 
 	if (command) scxfree(command);
     } else {
@@ -1035,11 +1042,9 @@ doext(char *command, double value)
 		    if ((cp = strchr(buff, '\n')))	/* contains newline */
 			*cp = '\0';			/* end string there */
 
-		    if (strlen(buff) + 1 > prevlen) {
-		    	prevlen = strlen(buff) + 40;
-			prevstr = scxrealloc(prevstr, prevlen);
-		    }
-		    (void) strcpy (prevstr, buff);
+		    if (!se->e.o.s || strlen(buff) != strlen(se->e.o.s))
+			se->e.o.s = scxrealloc(se->e.o.s, strlen(buff));
+		    (void) strcpy (se->e.o.s, buff);
 			 /* save alloc'd copy */
 		}
 		(void) pclose(pp);
@@ -1047,8 +1052,8 @@ doext(char *command, double value)
 	    } /* else */
 	} /* else */
     } /* else */
-    if (prevstr)
-	return (strcpy(scxmalloc((unsigned) (strlen(prevstr) + 1)), prevstr));
+    if (se->e.o.s)
+	return (strcpy(scxmalloc((unsigned) (strlen(se->e.o.s)+1)), se->e.o.s));
     else
 	return (strcpy(scxmalloc((unsigned)1), ""));
 }
@@ -1211,7 +1216,7 @@ seval(register struct enode *se)
 	    if (minc>maxc) c = maxc, maxc = minc, minc = c;
 	    return dostindex(minr, minc, maxr, maxc, se->e.o.right);
 	}
-	case EXT:    return (doext(seval(se->e.o.left), eval(se->e.o.right)));
+	case EXT:    return (doext(se));
 	case SVAL:   return (dosval(seval(se->e.o.left), eval(se->e.o.right)));
 	case SUBSTR: return (dosubstr(seval(se->e.o.left),
 			    (int)eval(se->e.o.right->e.o.left) - 1,
@@ -1280,7 +1285,6 @@ EvalAll()
 		error("Error in color 1: color turned off");
 	    }
 	}
-
     }
 
     (void) signal(SIGFPE, doquit);
@@ -1354,7 +1358,7 @@ RealEvalOne(register struct ent *p, int i, int j, int *chgct)
 	}
 	/* The following line is necessary to force gcc to clear the
 	 * FPU stack.  Otherwise v and p->v may not compare equal when
-	 * they really are due to comparing a double in memory with a
+	 * they really are, due to comparing a double in memory with a
 	 * higher precision value in the FPU, causing the dreaded
 	 * "Still changing after x iterations" message.  It may be
 	 * necessary for other compilers as well. - CRM
@@ -1383,6 +1387,7 @@ new(int op, struct enode *a1, struct enode *a2)
     p->op = op;
     p->e.o.left = a1;
     p->e.o.right = a2;
+    p->e.o.s = NULL;
     return p;
 }
 
@@ -1462,19 +1467,26 @@ copy(struct ent *dv1, struct ent *dv2, struct ent *v1, struct ent *v2)
     if (mindc>maxdc) c = maxdc, maxdc = mindc, mindc = c;
 
     if (v1 != NULL) {
-	maxsr = v2->row;
-	maxsc = v2->col;
 	minsr = v1->row;
 	minsc = v1->col;
+	maxsr = v2->row;
+	maxsc = v2->col;
 	if (minsr>maxsr) r = maxsr, maxsr = minsr, minsr = r;
 	if (minsc>maxsc) c = maxsc, maxsc = minsc, minsc = c;
+    } else if (showrange && !(showsr == currow && showsc == curcol &&
+	    mindr == currow && mindc == curcol &&
+	    maxdr == currow && maxdc == curcol)) {
+	minsr = showsr < currow ? showsr : currow;
+	minsc = showsc < curcol ? showsc : curcol;
+	maxsr = showsr > currow ? showsr : currow;
+	maxsc = showsc > curcol ? showsc : curcol;
     } else
 	if (minsr == -1) return;
 
     checkbounds(&maxdr, &maxdc);
 
-    flush_saved();
     erase_area(mindr, mindc, maxdr, maxdc);
+    flush_saved();
     if (minsr == maxsr && minsc == maxsc) {
 	/* Source is a single cell */
 	for (vr = mindr; vr <= maxdr; vr++)
@@ -1508,8 +1520,7 @@ copyrtv(int vr, int vc, int minsr, int minsc, int maxsr, int maxsc)
 	    if ((p = *ATBL(tbl, sr, sc))) {
 	    	n = lookat(dr, dc);
 		if (n->flags&is_locked) continue;
-		(void) clearent(n);
-		copyent(n, p, dr - sr, dc - sc);
+		copyent(n, p, dr - sr, dc - sc, 4, 0, maxrow, maxcol, 0);
 	    } else if ((n = *ATBL(tbl, dr, dc)))
 		(void) clearent(n);
 	}
@@ -1519,9 +1530,34 @@ copyrtv(int vr, int vc, int minsr, int minsc, int maxsr, int maxsc)
 void
 eraser(struct ent *v1, struct ent *v2)
 {
+    int i;
+    struct ent *obuf = NULL;
+
+    if (dbidx < 0) dbidx++;
+    delbuf[dbidx] = delbuf[DELBUFSIZE - 1];
+    delbuf[DELBUFSIZE - 1] = NULL;
+    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+	if (delbuf[i] == delbuf[dbidx]) {
+	    delbuf[dbidx] = NULL;
+	    break;
+	}
+    }
     flush_saved();
+    if (qbuf) {
+	if (dbidx < 0) dbidx++;
+	delbuf[dbidx] = delbuf[qbuf];
+	flush_saved();
+	obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+    }
     erase_area(v1->row, v1->col, v2->row, v2->col);
     sync_refs();
+    for (i = 0; i < DELBUFSIZE; i++)
+	if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+	    delbuf[i] = delbuf[dbidx];
+    qbuf = 0;
+    for (i = DELBUFSIZE - 1; i > DELBUFSIZE - 9; i--)
+	delbuf[i] = delbuf[i-1];
+    delbuf[DELBUFSIZE - 9] = delbuf[dbidx];
     FullUpdate++;
     modflg++;
 }
@@ -1530,15 +1566,40 @@ eraser(struct ent *v1, struct ent *v2)
 void
 yankr(struct ent *v1, struct ent *v2)
 {
+    int i, qtmp;
+    struct ent *obuf = NULL;
+
+    if (dbidx < 0) dbidx++;
+    delbuf[dbidx] = delbuf[DELBUFSIZE - 10];
+    delbuf[DELBUFSIZE - 10] = NULL;
+    for (i = dbidx + 1; i < DELBUFSIZE; i++) {
+	if (delbuf[i] == delbuf[dbidx]) {
+	    delbuf[dbidx] = NULL;
+	    break;
+	}
+    }
     flush_saved();
+    if (qbuf) {
+	if (dbidx < 0) dbidx++;
+	delbuf[dbidx] = delbuf[qbuf];
+	flush_saved();
+	obuf = delbuf[qbuf];	/* orig. contents of the del. buffer */
+    }
+    qtmp = qbuf;
+    qbuf = 0;
     yank_area(v1->row, v1->col, v2->row, v2->col);
+    qbuf = qtmp;
+    for (i = 0; i < DELBUFSIZE; i++)
+	if ((obuf && delbuf[i] == obuf) || (qbuf && i == qbuf))
+	    delbuf[i] = delbuf[dbidx];
+    qbuf = 0;
+    delbuf[DELBUFSIZE - 10] = delbuf[dbidx];
 }
 
 /* MOVE a Range of cells */
 void
 mover(struct ent *d, struct ent *v1, struct ent *v2)
 {
-    flush_saved();
     move_area(d->row, d->col, v1->row, v1->col, v2->row, v2->col);
     sync_refs();
     FullUpdate++;
@@ -1599,13 +1660,8 @@ moveto(int row, int col, int lastrow, int lastcol, int cornerrow,
 {
     register int i;
 
-    if (row != -1 && (row != currow || col != curcol)) {
-	/* Remember the current position */
-	savedrow[0] = currow;
-	savedcol[0] = curcol;
-	savedstrow[0] = strow;
-	savedstcol[0] = stcol;
-    }
+    if (!loading && row != -1 && (row != currow || col != curcol))
+	remember(0);
 
     currow = row;
     curcol = col;
@@ -1637,7 +1693,8 @@ moveto(int row, int col, int lastrow, int lastcol, int cornerrow,
     if (loading) {
 	update(1);
 	changed = 0;
-    }
+    } else
+	remember(1);
 }
 
 /*
@@ -1651,11 +1708,8 @@ num_search(double n, int firstrow, int firstcol, int lastrow,
     register int r,c;
     int	endr, endc;
 
-    /* Remember the current position */
-    savedrow[0] = currow;
-    savedcol[0] = curcol;
-    savedstrow[0] = strow;
-    savedstcol[0] = stcol;
+    if (!loading)
+	remember(0);
 
     g_free();
     gs.g_type = G_NUM;
@@ -1710,7 +1764,8 @@ num_search(double n, int firstrow, int firstcol, int lastrow,
     if (loading) {
 	update(1);
 	changed = 0;
-    }
+    } else
+	remember(1);
 }
 
 /* 'goto' a cell containing a matching string */
@@ -1723,11 +1778,8 @@ str_search(char *s, int firstrow, int firstcol, int lastrow, int lastcol,
     int		endr, endc;
     char	*tmp;
 
-    /* Remember the current position */
-    savedrow[0] = currow;
-    savedcol[0] = curcol;
-    savedstrow[0] = strow;
-    savedstcol[0] = stcol;
+    if (!loading)
+	remember(0);
 
 #if defined(RE_COMP)
     if ((tmp = re_comp(s)) != (char *)0) {
@@ -1847,7 +1899,8 @@ str_search(char *s, int firstrow, int firstcol, int lastrow, int lastcol,
     if (loading) {
 	update(1);
 	changed = 0;
-    }
+    } else
+	remember(1);
 }
 
 /* fill a range with constants */
@@ -1870,7 +1923,7 @@ fill(struct ent *v1, struct ent *v2, double start, double inc)
     if (minc < 0) minc = 0;
 
     FullUpdate++;
-    if( calc_order == BYROWS ) {
+    if (calc_order == BYROWS) {
     for (r = minr; r<=maxr; r++)
 	for (c = minc; c<=maxc; c++) {
 	    n = lookat(r, c);
@@ -1879,9 +1932,10 @@ fill(struct ent *v1, struct ent *v2, double start, double inc)
 	    n->v = start;
 	    start += inc;
 	    n->flags |= (is_changed|is_valid);
+	    n->flags &= ~(is_cleared);
 	}
     }
-    else if ( calc_order == BYCOLS ) {
+    else if (calc_order == BYCOLS) {
     for (c = minc; c<=maxc; c++)
 	for (r = minr; r<=maxr; r++) {
 	    n = lookat(r, c);
@@ -1889,6 +1943,7 @@ fill(struct ent *v1, struct ent *v2, double start, double inc)
 	    n->v = start;
 	    start += inc;
 	    n->flags |= (is_changed|is_valid);
+	    n->flags &= ~(is_cleared);
 	}
     }
     else error(" Internal error calc_order");
@@ -2013,6 +2068,22 @@ let(struct ent *v, struct enode *e)
 
     changed++; modflg++;
     v->flags |= (is_changed|is_valid);
+    
+    if (!loading) {
+	int i;
+
+	for (i = 36; i > 28; i--) {
+	    savedrow[i] = savedrow[i-1];
+	    savedcol[i] = savedcol[i-1];
+	    savedstrow[i] = savedstrow[i-1];
+	    savedstcol[i] = savedstcol[i-1];
+	}
+
+	savedrow[28] = v->row;
+	savedcol[28] = v->col;
+	savedstrow[28] = savedstrow[27];
+	savedstcol[28] = savedstcol[27];
+    }
 }
 
 void
@@ -2066,6 +2137,22 @@ slet(struct ent *v, struct enode *se, int flushdir)
 	v->flags |= is_label;
     else v->flags &= ~is_label;
 
+    if (!loading) {
+	int i;
+
+	for (i = 36; i > 28; i--) {
+	    savedrow[i] = savedrow[i-1];
+	    savedcol[i] = savedcol[i-1];
+	    savedstrow[i] = savedstrow[i-1];
+	    savedstcol[i] = savedstcol[i-1];
+	}
+
+	savedrow[28] = v->row;
+	savedcol[28] = v->col;
+	savedstrow[28] = savedstrow[27];
+	savedstcol[28] = savedstcol[27];
+    }
+
     FullUpdate++;
     changed++;
     modflg++;
@@ -2153,9 +2240,7 @@ clearent(struct ent *v)
     if (v->format)
 	scxfree(v->format);
     v->format = (char *)0;
-    v->flags |= (is_changed);
-    v->flags &= ~(is_valid);
-    v->flags |= (is_cleared);
+    v->flags = (is_changed|is_cleared);
     changed++;
     modflg++;
 }
@@ -2199,6 +2284,8 @@ efree(struct enode *e)
 	}
 	if (e->op == O_SCONST && e->e.s)
 	    scxfree(e->e.s);
+	else if (e->op == EXT && e->e.o.s)
+	    scxfree(e->e.o.s);
 	e->e.o.left = freeenodes;
 	freeenodes = e;
     }
@@ -2467,7 +2554,7 @@ list_arg(char *s, struct enode *e)
     for (; (line[linelim++] = *s++); );
     linelim--;
 
-    decompile (e->e.o.right, 0);
+    decompile(e->e.o.right, 0);
     line[linelim++] = ',';
     decompile_list(e->e.o.left);
     line[linelim - 1] = ')';
@@ -2478,7 +2565,7 @@ one_arg(char *s, struct enode *e)
 {
     for (; (line[linelim++] = *s++); );
     linelim--;
-    decompile (e->e.o.right, 0);
+    decompile(e->e.o.right, 0);
     line[linelim++] = ')';
 }
 
@@ -2487,7 +2574,7 @@ two_arg(char *s, struct enode *e)
 {
     for (; (line[linelim++] = *s++); );
     linelim--;
-    decompile (e->e.o.left, 0);
+    decompile(e->e.o.left, 0);
     line[linelim++] = ',';
     decompile (e->e.o.right, 0);
     line[linelim++] = ')';
@@ -2498,9 +2585,9 @@ three_arg(char *s, struct enode *e)
 {
     for (; (line[linelim++] = *s++); );
     linelim--;
-    decompile (e->e.o.left, 0);
+    decompile(e->e.o.left, 0);
     line[linelim++] = ',';
-    decompile (e->e.o.right->e.o.left, 0);
+    decompile(e->e.o.right->e.o.left, 0);
     line[linelim++] = ',';
     decompile (e->e.o.right->e.o.right, 0);
     line[linelim++] = ')';
@@ -2545,11 +2632,12 @@ editv(int row, int col)
     p = lookat(row, col);
     (void) sprintf(line, "let %s = ", v_name(row, col));
     linelim = strlen(line);
-    if ((p->flags & is_strexpr || p->expr == NULL) && p->flags & is_valid) {
-	(void) sprintf(line+linelim, "%.15g", p->v);
-	linelim = strlen(line);
-    } else if (p->expr) {
-        editexp(row, col);
+    if (p->flags & is_valid) {
+	if (p->flags & is_strexpr || p->expr == NULL) {
+	    (void) sprintf(line+linelim, "%.15g", p->v);
+	    linelim = strlen(line);
+	} else
+	    editexp(row, col);
     }
 }
 
