@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3/88, see list of changes.
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  *
  */
 
@@ -175,8 +175,7 @@ yylex()
 	    if (ret == WORD) {
 		struct range *r;
 		if (!find_range(tokenst, tokenl,
-			(struct ent *)0, (struct ent *)0, &r) &&
-			strlen(r->r_name) == tokenl) {
+			(struct ent *)0, (struct ent *)0, &r)) {
 		    yylval.rval.left = r->r_left;
 		    yylval.rval.right = r->r_right;
 		    if (r->r_is_range)
