@@ -1,10 +1,10 @@
 /*	SC	A Spreadsheet Calculator
  *		Routines for piping data to and from an external macro program
  *
- *		Chuck Martin <cmartin@bigfoot.com>
+ *		Chuck Martin <nrocinu@myrealbox.com>
  *		Original Version Created:  June, 2000
  *
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  */
 
 #include <curses.h>
