@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  *
  */
 
@@ -53,8 +53,8 @@ struct ent ***tbl;
 int arg = 1;
 int strow = 0, stcol = 0;
 int currow = 0, curcol = 0;
-int savedrow[27], savedcol[27];
-int savedstrow[27], savedstcol[27];
+int savedrow[37], savedcol[37];
+int savedstrow[37], savedstcol[37];
 int FullUpdate = 0;
 int maxrow, maxcol;
 int maxrows, maxcols;
@@ -65,13 +65,15 @@ char *col_hidden;
 char *row_hidden;
 char line[FBUFLEN];
 int changed;
-struct ent *delbuf[4];
+struct ent *delbuf[DELBUFSIZE];
 int dbidx;
+int qbuf;	/* buffer no specified by " command */
 int modflg;
 int cellassign;
 int numeric;
 char *mdir;
 char *autorun;
+int skipautorun;
 char *fkey[FKEYS];
 char *scext;
 char *ascext;
@@ -103,6 +105,8 @@ int  showneed  = 0;	/* Causes cells needing values to be highlighted  */
 int  showexpr  = 0;	/* Causes cell exprs to be displayed, highlighted */
 int  shownote  = 0;	/* Causes cells with attached notes to be
 			   highlighted					  */
+int  braille   = 0;	/* Be nice to users of braille displays		  */
+int  braillealt = 0;	/* Alternate mode for braille users		  */
 
 int  autocalc  = 1;	/* 1 to calculate after each update */
 int  autolabel = 1;     /* If room, causes label to be created after a define */
@@ -154,6 +158,7 @@ lookat(int row, int col)
         if (freeents != NULL) {
 	    *pp = freeents;
 	    (*pp)->flags &= ~(is_cleared);
+	    (*pp)->flags |= may_sync;
 	    freeents = freeents->next;
 	} else
 	    *pp = (struct ent *) scxmalloc((unsigned)sizeof(struct ent));
@@ -164,7 +169,7 @@ lookat(int row, int col)
 	(*pp)->col = col;
 	(*pp)->nrow = -1;
 	(*pp)->ncol = -1;
-	(*pp)->flags = 0;
+	(*pp)->flags = may_sync;
 	(*pp)->expr = (struct enode *)0;
 	(*pp)->v = (double) 0.0;
 	(*pp)->format = (char *)0;
@@ -231,14 +236,16 @@ main (int argc, char  **argv)
      * command line overrides file options
      */
 
-    int Mopt = 0;
-    int Oopt = 0;
-    int Nopt = 0;
+    int mopt = 0;
+    int oopt = 0;
+    int nopt = 0;
+    int copt = 0; 
+    int ropt = 0;
     int Copt = 0; 
     int Ropt = 0;
-    int Eopt = 0;
-    int Popt = 0;
-    int Qopt = 0;
+    int eopt = 0;
+    int popt = 0;
+    int qopt = 0;
 
     Vopt = 0;
 
@@ -255,8 +262,11 @@ main (int argc, char  **argv)
     else
 	progname = argv[0];
 
-    while ((c = getopt(argc, argv, "xmoncrCReP:vq")) != EOF) {
+    while ((c = getopt(argc, argv, "axmoncrCReP:vq")) != EOF) {
     	switch (c) {
+	    case 'a':
+		    skipautorun = 1;
+		    break;
 	    case 'x':
 #if defined(VMS) || defined(MSDOS) || !defined(CRYPT_PATH)
 		    (void) fprintf(stderr, "Crypt not available\n");
@@ -266,45 +276,47 @@ main (int argc, char  **argv)
 #endif
 		    break;
 	    case 'm':
-		    Mopt = 1;
+		    mopt = 1;
 		    break;
 	    case 'o':
-		    Oopt = 1;
+		    oopt = 1;
 		    break;
 	    case 'n':
-		    Nopt = 1;
+		    nopt = 1;
 		    break;
 	    case 'c':
-		    Copt = 1;
+		    copt = 1;
 		    break;
 	    case 'r':
-		    Ropt = 1;
+		    ropt = 1;
 		    break;
 	    case 'C':
+		    Copt = 1;
 		    craction = CRCOLS;
 		    break;
 	    case 'R':
+		    Ropt = 1;
 		    craction = CRROWS;
 		    break;
 	    case 'e':
 		    rndtoeven = 1;
-		    Eopt = 1;
+		    eopt = 1;
 		    break;
 	    case 'P':
-		    Popt = 1;
+		    popt = 1;
 	    case 'v':
 		    break;
 	    case 'q':
-		    Qopt = 1;
+		    qopt = 1;
 		    break;
 	    default:
-		    (void) fprintf(stderr,"Usage: %s [-cemnoqrxCR] [file...]\n",
+		    (void) fprintf(stderr,"Usage: %s [-acemnoqrxCR] [file...]\n",
 			progname);
 		    exit (1);
 	}
     }
 
-    if (!isatty(STDOUT_FILENO) || Popt || Qopt) usecurses = FALSE;
+    if (!isatty(STDOUT_FILENO) || popt || qopt) usecurses = FALSE;
     startdisp();
     signals();
 
@@ -318,7 +330,7 @@ main (int argc, char  **argv)
      * Build revision message for later use:
      */
 
-    if (Popt)
+    if (popt)
 	*revmsg = '\0';
     else {
 	(void) strcpy(revmsg, progname);
@@ -328,15 +340,17 @@ main (int argc, char  **argv)
 	(void) strcat(revmsg, ":  Type '?' for help.");
     }
 
-#ifndef MSDOS
+#ifdef MSDOS
+    if (optind < argc)
+#else 
     if (optind < argc && !strcmp(argv[optind], "--"))
 	optind++;
     if (optind < argc && argv[optind][0] != '|' &&
 	    !strcmp(argv[optind], "-"))
-#else 
-    if (optind < argc)
 #endif /* MSDOS */
 	(void) strcpy(curfile, argv[optind]);
+    for (dbidx = DELBUFSIZE - 1; dbidx >= 0; )
+	delbuf[dbidx--] = NULL;
     if (optind < argc) {
 	readfile(argv[optind], 1);
 	EvalAll();
@@ -347,12 +361,16 @@ main (int argc, char  **argv)
 	readfile(argv[optind], 0);
 	optind++;
     }
+    savedrow[0] = currow;
+    savedcol[0] = curcol;
+    savedstrow[0] = strow;
+    savedstcol[0] = stcol;
     EvalAll();
 
-    if (!(Popt || isatty(STDIN_FILENO)))
+    if (!(popt || isatty(STDIN_FILENO)))
 	readfile("-", 0);
 
-    if (Qopt) {
+    if (qopt) {
 	stopdisp();
 	exit (0);
     }
@@ -363,19 +381,23 @@ main (int argc, char  **argv)
     clearok(stdscr, TRUE);
     EvalAll();
 
-    if (Mopt)
+    if (mopt)
 	autocalc = 0;
-    if (Oopt)
+    if (oopt)
 	optimize = 1;
-    if (Nopt)
+    if (nopt)
 	numeric = 1;
-    if (Copt)
+    if (copt)
 	calc_order = BYCOLS;
-    if (Ropt)
+    if (ropt)
 	calc_order = BYROWS;
-    if (Eopt)
+    if (Copt)
+	craction = CRCOLS;
+    if (Ropt)
+	craction = CRROWS;
+    if (eopt)
 	rndtoeven = 1;
-    if (Popt) {
+    if (popt) {
 	char *redraw = NULL;
 	int o;
 
@@ -446,7 +468,7 @@ main (int argc, char  **argv)
 			     */
 			    if (mode_ind == 'i')
 				write_line(ctl('v'));
-			    error("Select range:");
+			    error("");
 			    update(1);
 			}
 			stopdisp();
@@ -752,6 +774,7 @@ main (int argc, char  **argv)
 		    error(
 "Toggle: a:auto,c:cell,e:ext funcs,n:numeric,t:top,$:pre-scale,<MORE>");
 #endif
+		    if (braille) move(1, 0);
 		    (void) refresh();
 
 		    switch (nmgetch()) {
@@ -780,6 +803,13 @@ main (int argc, char  **argv)
 			    repaint(lastmx, lastmy, fwidth[lastcol]);
 			    error("Cell highlighting %sabled.",
 				    showcell ? "en" : "dis");
+			    --modflg;	/* negate the modflg++ */
+			    break;
+			case 'b':
+			    braille ^= 1;
+			    error("Braille enhancement %sabled.",
+				    braille ? "en" : "dis");
+			    --modflg;	/* negate the modflg++ */
 			    break;
 			case 's':
 			    cslop ^= 1;
@@ -837,6 +867,7 @@ main (int argc, char  **argv)
 			    break;
 			case ESC:
 			case ctl('g'):
+			    error("");
 			    --modflg;	/* negate the modflg++ */
 			    break;
 			case 'r': case 'R':
@@ -848,11 +879,13 @@ main (int argc, char  **argv)
 				    break;
 				case 'j':
 				case ctl('n'):
+				case KEY_DOWN:
 				    craction = CRROWS;
 				    error("Down row after new line");
 				    break;
 				case 'l':
 				case ' ':
+				case KEY_RIGHT:
 				    craction = CRCOLS;
 				    error("Right column after new line");
 				    break;
@@ -921,16 +954,12 @@ main (int argc, char  **argv)
 		    if (linelim >= 0)
 			write_line(c);
 		    else {
-			/* Remember the current position */
-			savedrow[0] = currow;
-			savedcol[0] = curcol;
-			savedstrow[0] = strow;
-			savedstcol[0] = stcol;
-
+			remember(0);
 			currow = 0;
 			curcol = 0;
 			rowsinrange = 1;
 			colsinrange = fwidth[curcol];
+			remember(1);
 			FullUpdate++;
 		    }
 		    break;
@@ -946,15 +975,11 @@ main (int argc, char  **argv)
 		    if (linelim >= 0)
 			write_line(c);
 		    else {
-			/* Remember the current position */
-			savedrow[0] = currow;
-			savedcol[0] = curcol;
-			savedstrow[0] = strow;
-			savedstcol[0] = stcol;
-
+			remember(0);
 			curcol = 0;
 			rowsinrange = 1;
 			colsinrange = fwidth[curcol];
+			remember(1);
 		    }
 		} else {
 		    nedistate = 0;
@@ -1038,21 +1063,36 @@ main (int argc, char  **argv)
 		case '.':
 		    if (locked_cell(currow, curcol))
 			break;
+		    /* set mark 0 */
+		    savedrow[27] = currow;
+		    savedcol[27] = curcol;
+		    savedstrow[27] = strow;
+		    savedstcol[27] = stcol;
+
 		    numeric_field = 1;
 		    (void) sprintf(line,"let %s = %c",
 			    v_name(currow, curcol), c);
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 
 		case '+':
 		case '-':
 		    if (!locked_cell(currow, curcol)) {
+			struct ent *p = lookat(currow, curcol);
+			/* set mark 0 */
+			savedrow[27] = currow;
+			savedcol[27] = curcol;
+			savedstrow[27] = strow;
+			savedstcol[27] = stcol;
+
 			numeric_field = 1;
 			editv(currow, curcol);
-			edit_mode();
-			write_line('A');
+			insert_mode();
+			if (p->flags & is_valid) {
+			    write_line(ESC);
+			    write_line('a');
+			}
 			write_line(c);
 		    }
 		    break;
@@ -1060,10 +1100,15 @@ main (int argc, char  **argv)
 		case '=':
 		    if (locked_cell(currow, curcol))
 			break;
+		    /* set mark 0 */
+		    savedrow[27] = currow;
+		    savedcol[27] = curcol;
+		    savedstrow[27] = strow;
+		    savedstcol[27] = stcol;
+
 		    (void) sprintf(line,"let %s = ", v_name(currow, curcol));
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 
 		case '!':
@@ -1124,112 +1169,125 @@ main (int argc, char  **argv)
 		 */
 
 		case '/':
-		    edit_mode();
 		    error(
 "Range: x:erase v:value c:copy f:fill d:def l:lock U:unlock S:show u:undef F:fmt");
+		    if (braille) move(1, 0);
 		    (void) refresh();
 
-		    switch (nmgetch()) {
+		    c = nmgetch();
+		    error("");
+		    switch (c) {
 		    case 'l':
 			(void) sprintf(line,"lock [range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'U':
 			(void) sprintf(line,"unlock [range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'c':
 			(void) sprintf(line,"copy [dest_range src_range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'm':
-			error("");
 			(void) sprintf(line,"move [destination src_range] %s ",
-			    v_name(currow, curcol));
-			write_line('A');
+				v_name(currow, curcol));
+			linelim = strlen(line);
+			insert_mode();
 		        write_line(ctl('v'));
 			break;
 		    case 'x':
 			(void) sprintf(line,"erase [range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'y':
 			(void) sprintf(line,"yank [range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'v':
 			(void) sprintf(line, "value [range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'f':
 			(void) sprintf(line,"fill [range start inc] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'd':
 			(void) sprintf(line,"define [string range] \"");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			break;
 		    case 'u':
 			(void) sprintf(line,"undefine [range] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			break;
 		    case 'r':
 			error("frame (top/bottom/left/right/all/unframe)");
+			if (braille) move(1, 0);
 			refresh();
 			linelim = 0;
-			switch (nmgetch()) {
-			case 't':
-			    sprintf(line,"frametop [<outrange> rows] ");
-			    break;
-			case 'b':
-			    sprintf(line,"framebottom [<outrange> rows] ");
-			    break;
-			case 'l':
-			    sprintf(line,"frameleft [<outrange> cols] ");
-			    break;
-			case 'r':
-			    sprintf(line,"frameright [<outrange> cols] ");
-			    break;
-			case 'a':
-			    sprintf(line,"frame [<outrange> inrange] ");
-			    write_line('A');
-			    startshow();
-			    break;
-			case 'u':
-			    sprintf(line,"unframe [<range>] ");
-			    write_line('A');
-			    startshow();
-			    break;
-			case ESC:
-			case ctl('g'):
-			    error("");
-			    linelim = -1;
-			    break;
-			default:
-			    error("Invalid frame command");
-			    linelim = -1;
-			    break;
+			c = nmgetch();
+			error("");
+			switch (c) {
+			    case 't':
+				sprintf(line,"frametop [<outrange> rows] ");
+				break;
+			    case 'b':
+				sprintf(line,"framebottom [<outrange> rows] ");
+				break;
+			    case 'l':
+				sprintf(line,"frameleft [<outrange> cols] ");
+				break;
+			    case 'r':
+				sprintf(line,"frameright [<outrange> cols] ");
+				break;
+			    case 'a':
+				sprintf(line,"frame [<outrange> inrange] ");
+				break;
+			    case 'u':
+				sprintf(line,"unframe [<range>] ");
+				break;
+			    case ESC:
+			    case ctl('g'):
+				linelim = -1;
+				break;
+			    default:
+				error("Invalid frame command");
+				linelim = -1;
+				break;
 			}
 			if (linelim == 0) {
-			    error("");
-			    write_line('A');
+			    linelim = strlen(line);
+			    insert_mode();
 			}
+			if (c == 'a' || c == 'u')
+			    startshow();
 			break;
 		    case 's':
 			(void) sprintf(line, "sort [range \"criteria\"] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'C':
 			(void) sprintf(line,"color [range color#] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case 'S':	/* Show color definitions and ranges */
@@ -1263,12 +1321,12 @@ main (int argc, char  **argv)
 			break;
 		    case 'F':
 			(void) sprintf(line, "fmt [range \"format\"] ");
-			write_line('A');
+			linelim = strlen(line);
+			insert_mode();
 			startshow();
 			break;
 		    case ESC:
 		    case ctl('g'):
-			error("");
 			break;
 		    default:
 			error("Invalid region command");
@@ -1276,6 +1334,23 @@ main (int argc, char  **argv)
 		    }
 		    break;
 
+		case '"':
+		    error("Select buffer (0-9 or a-z):");
+		    if ((c=nmgetch()) == ESC || c == ctl('g')) {
+			error("");
+		    } else if (c >= '0' && c <= '9') {
+			qbuf = c - '0' + (DELBUFSIZE - 10);
+			error("");
+		    } else if (c >= 'a' && c <= 'z') {
+			qbuf = c - 'a' + (DELBUFSIZE - 36);
+			error("");
+		    } else if (c == '"') {
+			qbuf = 0;
+			error("");
+		    } else
+			error("Invalid buffer");
+		    break;
+
 		/*
 		 * Row/column commands:
 		 */
@@ -1385,34 +1460,26 @@ main (int argc, char  **argv)
 		    {
 		    register struct ent *p;
 
-		    /* Remember the current position */
-		    savedrow[0] = currow;
-		    savedcol[0] = curcol;
-		    savedstrow[0] = strow;
-		    savedstcol[0] = stcol;
-
+		    remember(0);
 		    curcol = maxcols - 1;
 		    while (!VALID_CELL(p, currow, curcol) && curcol > 0)
 			curcol--;
 		    rowsinrange = 1;
 		    colsinrange = fwidth[curcol];
+		    remember(1);
 		    break;
 		    }
 		case '#':
 		    {
 		    register struct ent *p;
 
-		    /* Remember the current position */
-		    savedrow[0] = currow;
-		    savedcol[0] = curcol;
-		    savedstrow[0] = strow;
-		    savedstcol[0] = stcol;
-
+		    remember(0);
 		    currow = maxrows - 1;
 		    while (!VALID_CELL(p, currow, curcol) && currow > 0)
 			currow--;
 		    rowsinrange = 1;
 		    colsinrange = fwidth[curcol];
+		    remember(1);
 		    break;
 		    }
 		case 'w':
@@ -1467,15 +1534,11 @@ main (int argc, char  **argv)
 		    break;
 		    }
 		case '^':
-		    /* Remember the current position */
-		    savedrow[0] = currow;
-		    savedcol[0] = curcol;
-		    savedstrow[0] = strow;
-		    savedstcol[0] = stcol;
-
+		    remember(0);
 		    currow = 0;
 		    rowsinrange = 1;
 		    colsinrange = fwidth[curcol];
+		    remember(1);
 		    break;
 #ifdef KEY_HELP
 		case KEY_HELP:
@@ -1485,31 +1548,46 @@ main (int argc, char  **argv)
 		    break;
 		case '\\':
 		    if (!locked_cell(currow, curcol)) {
+			/* set mark 0 */
+			savedrow[27] = currow;
+			savedcol[27] = curcol;
+			savedstrow[27] = strow;
+			savedstcol[27] = stcol;
+
 			(void) sprintf(line, "label %s = \"",
 				v_name(currow, curcol));
 			linelim = strlen(line);
-			edit_mode();
-			write_line('A');
+			insert_mode();
 		    }
 		    break;
 
 		case '<':
 		    if (!locked_cell(currow, curcol)) {
+			/* set mark 0 */
+			savedrow[27] = currow;
+			savedcol[27] = curcol;
+			savedstrow[27] = strow;
+			savedstcol[27] = stcol;
+
 			(void) sprintf(line, "leftstring %s = \"",
 				v_name(currow, curcol));
 			linelim = strlen(line);
-			edit_mode();
-			write_line('A');
+			insert_mode();
 		    }
 		    break;
 
 		case '>':
 		    if (!locked_cell(currow, curcol)) {
+			/* set mark 0 */
+			savedrow[27] = currow;
+			savedcol[27] = curcol;
+			savedstrow[27] = strow;
+			savedstcol[27] = stcol;
+
 		       (void) sprintf(line, "rightstring %s = \"",
 			      v_name(currow, curcol));
 		       linelim = strlen(line);
-		       edit_mode();
-		       write_line('A');
+		       insert_mode();
 		    }
 		    break;
 		case '{':
@@ -1551,50 +1629,56 @@ main (int argc, char  **argv)
 		case 'e':
 		    if (!locked_cell(currow, curcol)) {
 			struct ent *p = lookat(currow, curcol);
+
+			/* set mark 0 */
+			savedrow[27] = currow;
+			savedcol[27] = curcol;
+			savedstrow[27] = strow;
+			savedstcol[27] = stcol;
+
 			editv(currow, curcol);
-			edit_mode();
-			if (!(p->flags & is_valid))
-			    write_line('A');
+			if (!(p->flags & is_valid)) {
+			    linelim = strlen(line);
+			    insert_mode();
+			} else
+			    edit_mode();
 		    }
 		    break;
 		case 'E':
 		    if (!locked_cell(currow, curcol)) {
+			/* set mark 0 */
+			savedrow[27] = currow;
+			savedcol[27] = curcol;
+			savedstrow[27] = strow;
+			savedstcol[27] = stcol;
+
 			edits(currow, curcol);
 			edit_mode();
 		    }
 		    break;
 		case 'f':
-		    if (arg == 1)
-			(void) sprintf(line, "format [for column] %s ",
-				coltoa(curcol));
-		    else {
-			(void) sprintf(line, "format [for columns] %s:",
-				coltoa(curcol));
-			(void) sprintf(line+strlen(line), "%s ",
-				coltoa(curcol+arg-1));
-		    }
-		    error("Current format is %d %d %d",
-			fwidth[curcol],precision[curcol],realfmt[curcol]);
-		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    formatcol(arg);
 		    break;
 		case 'F': {
 		    register struct ent *p = *ATBL(tbl, currow, curcol);
 		    if (p && p->format) {
 			(void) sprintf(line, "fmt [format] %s \"%s",
-			    v_name(currow, curcol), p->format);
+				v_name(currow, curcol), p->format);
 			edit_mode();
 			linelim = strlen(line) - 1;
 		    } else {
 			(void) sprintf(line, "fmt [format] %s \"",
 				   v_name(currow, curcol));
-			edit_mode();
-			write_line('A');
+			insert_mode();
+			linelim = strlen(line);
 		    }
 		    break;
 		}
 		case 'C': {
+		    if (braille) {
+			braillealt ^= 1;
+			break;
+		    }
 		    error("Color number to set (1-8)?");
 		    if ((c=nmgetch()) == ESC || c == ctl('g')) {
 			error("");
@@ -1612,8 +1696,7 @@ main (int argc, char  **argv)
 			line[linelim] = '\0';
 			edit_mode();
 		    } else {
-			edit_mode();
-			write_line('A');
+			insert_mode();
 		    }
 		    break;
 		}
@@ -1623,8 +1706,7 @@ main (int argc, char  **argv)
 		case 'g':
 		    (void) sprintf(line, "goto [v] ");
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'P':
 		    (void) sprintf(line, "put [\"dest\" range] \"");
@@ -1653,14 +1735,12 @@ main (int argc, char  **argv)
 			    strlen(curfile + strlen(curfile) + 1)) = '\0';
 		    curfile[strlen(curfile)] = c;
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'M':
 		    (void) sprintf(line, "merge [\"source\"] \"");
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'R':
 		    if (mdir)
@@ -1668,14 +1748,12 @@ main (int argc, char  **argv)
 		    else
 			(void) sprintf(line,"merge [\"macro_file\"] \"");
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'D':
 		    (void) sprintf(line, "mdir [\"macro_directory\"] \"");
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'A':
 		    if (autorun)
@@ -1683,16 +1761,14 @@ main (int argc, char  **argv)
 		    else
 			(void) sprintf(line, "autorun [\"macro_file\"] \"");
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'G':
 		    (void) sprintf(line, "get [\"source\"] \"");
 		    if (*curfile)
 			error("Default file is \"%s\"", curfile);
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'W':
 		    (void) sprintf(line, "write [\"dest\" range] \"");
@@ -1735,15 +1811,13 @@ main (int argc, char  **argv)
 			    strlen(curfile + strlen(curfile) + 1)) = '\0';
 		    curfile[strlen(curfile)] = c;
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'S':	/* set options */
 		    (void) sprintf (line, "set ");
 		    error("Options:byrows,bycols,iterations=n,tblstyle=(0|tbl|latex|slatex|tex|frame),<MORE>");
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 		case 'T':	/* tbl output */
 		    (void) sprintf (line, "tbl [\"dest\" range] \"");
@@ -1784,33 +1858,18 @@ main (int argc, char  **argv)
 			    strlen(curfile + strlen(curfile) + 1)) = '\0';
 		    curfile[strlen(curfile)] = c;
 		    linelim = strlen(line);
-		    edit_mode();
-		    write_line('A');
+		    insert_mode();
 		    break;
 #ifdef KEY_DC
 		case KEY_DC:
 #endif
 		case 'x':
-		    {
-		    register struct ent **pp;
-		    register int c1;
-
-		    flush_saved();
-		    if (arg == 1) {
-			pp = ATBL(tbl, currow, curcol);
-			if ((*pp) && !locked_cell(currow, curcol)) {
-			    dbidx++;
-			    free_ent(*pp);
-			    *pp = NULL;
-			}
-		    } else if (calc_order == BYROWS)
-			erase_area(currow, curcol, currow, curcol + arg - 1);
+		    if (calc_order == BYROWS)
+			eraser(lookat(currow, curcol),
+				lookat(currow, curcol + arg - 1));
 		    else
-			erase_area(currow, curcol, currow + arg - 1, curcol);
-		    sync_refs();
-		    modflg++;
-		    FullUpdate++;
-		    }
+			eraser(lookat(currow, curcol),
+				lookat(currow + arg - 1, curcol));
 		    break;
 		case 'Q':
 		case 'q':
@@ -1860,16 +1919,12 @@ main (int argc, char  **argv)
 		    break;
 #ifdef KEY_HOME
 		case KEY_HOME:
-		    /* Remember the current position */
-		    savedrow[0] = currow;
-		    savedcol[0] = curcol;
-		    savedstrow[0] = strow;
-		    savedstcol[0] = stcol;
-
+		    remember(0);
 		    currow = 0;
 		    curcol = 0;
 		    rowsinrange = 1;
 		    colsinrange = fwidth[curcol];
+		    remember(1);
 		    FullUpdate++;
 		    break;
 #endif
@@ -1890,8 +1945,11 @@ main (int argc, char  **argv)
 			error("");
 			break;
 		    }
-		    if ((c -= ('a' - 1)) < 1 || c > 26) {
-			error("Invalid character for mark (must be a-z)");
+		    if (c == '`' || c == '\'')
+			c = 0;
+		    else if (!(((c -= ('a' - 1)) > 0 && c < 27) ||
+			    ((c += ('a' - '0' + 26)) > 26 && c < 37))) {
+			error("Invalid mark (must be a-z, 0-9, ` or \')");
 			break;
 		    }
 		    if (savedrow[c] == -1) {
@@ -1910,7 +1968,8 @@ main (int argc, char  **argv)
 			for (c1 = curcol; arg-- && c1 < maxcols; c1++) {
 			    n = lookat(currow, c1);
 			    (void) clearent(n);
-			    copyent(n, p, currow - savedrow[c], c1 - savedcol[c]);
+			    copyent(n, p, currow - savedrow[c],
+				    c1 - savedcol[c], 0, 0, maxrow, maxcol, 0);
 			}
 		    break;
 		    }
@@ -1931,8 +1990,7 @@ main (int argc, char  **argv)
 			    sprintf(line, "addnote [target range] %s ", 
 				    v_name(currow, curcol));
 			    linelim = strlen(line);
-			    edit_mode();
-			    write_line('A');
+			    insert_mode();
 			    write_line(ctl('v'));
 			    error("");
 			    break;
@@ -1987,6 +2045,11 @@ main (int argc, char  **argv)
 			    break;
 		    }
 		    break;
+		case KEY_RESIZE:
+#ifndef	SIGWINCH
+		    winchg();
+#endif
+		    break;
 		default:
 		    if ((toascii(c)) != c)
 			error ("Weird character, decimal %d\n",
@@ -2068,7 +2131,6 @@ nopipe()
     brokenpipe = TRUE;
 }
 
-#ifdef	SIGWINCH
 #ifdef SIGVOID
 void
 #else
@@ -2078,14 +2140,20 @@ winchg()
 {
     stopdisp();
     startdisp();
+    /*
+     * I'm not sure why a refresh() needs to be done both before and after
+     * the clearok() and update(), but without doing it this way, a screen
+     * (or window) that grows bigger will leave the added space blank. - CRM
+     */
     refresh();
     FullUpdate++;
     (void) clearok(stdscr, TRUE);
     update(1);
     refresh();
+#ifdef	SIGWINCH
     (void) signal(SIGWINCH, winchg);
-}
 #endif
+}
 
 #ifdef SIGVOID
 void
