@@ -6,7 +6,7 @@
  *			University of Maryland
  *		R. Bond  12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  *
  */
 
@@ -39,6 +39,7 @@
 #define HISTLEN		10	/* Number of history entries for vi emulation */
 #define CPAIRS		 8	/* Number of color pairs available */
 #define COLFORMATS	10	/* Number of custom column formats */
+#define DELBUFSIZE	40	/* Number of named buffers + 4 */
 #ifdef PSC
 # define error(msg)	fprintf(stderr, msg);
 #else
@@ -168,7 +169,8 @@ struct enode {
 	char *s;		/* string part of a cell */
 	struct {		/* other cells use to eval()/seval() */
 	    struct enode *left, *right;
-	} o;
+	    char *s;		/* previous value of @ext function in case */
+	} o;			/*	external functions are turned off */
     } e;
 };
 
@@ -274,6 +276,7 @@ struct go_save {
 #define is_locked    0040
 #define is_label     0100
 #define is_cleared   0200
+#define may_sync     0400
 
 /* cell error (1st generation (ERROR) or 2nd+ (INVALID)) */
 #define	CELLOK		0
@@ -313,8 +316,8 @@ extern	int strow, stcol;
 extern	int currow, curcol;
 extern	int gmyrow, gmycol;	/* globals used for @myrow, @mycol cmds */
 extern	int rescol;		/* columns reserved for row numbers */
-extern	int savedrow[27], savedcol[27];
-extern	int savedstrow[27], savedstcol[27];
+extern	int savedrow[37], savedcol[37];
+extern	int savedstrow[37], savedstcol[37];
 extern	int FullUpdate;
 extern	int maxrow, maxcol;
 extern	int maxrows, maxcols;	/* # cells currently allocated */
@@ -329,8 +332,9 @@ extern	char *row_hidden;
 extern	char line[FBUFLEN];
 extern	int linelim;
 extern	int changed;
-extern	struct ent *delbuf[4];
+extern	struct ent *delbuf[DELBUFSIZE];
 extern	int dbidx;
+extern	int qbuf;	/* buffer no specified by " command */
 extern	int showsc, showsr;
 extern	int showrange;		/* Causes ranges to be highlighted */
 extern	int cellassign;
@@ -370,7 +374,8 @@ extern	int nmgetch();
 extern	int writefile(char *fname, int r0, int c0, int rn, int cn);
 extern	int yn_ask(char *msg);
 extern	struct colorpair *cpairs[8];
-extern	struct enode *copye(register struct enode *e, int Rdelta, int Cdelta);
+extern	struct enode *copye(register struct enode *e, int Rdelta, int Cdelta,
+	int r1, int c1, int r2, int c2, int transpose);
 extern	struct enode *new(int op, struct enode *a1, struct enode *a2);
 extern	struct enode *new_const(int op, double a1);
 extern	struct enode *new_range(int op, struct range_s a1);
@@ -400,8 +405,8 @@ extern	void closerow(int r, int numrow);
 extern	void colshow_op();
 extern	void copy(struct ent *dv1, struct ent *dv2, struct ent *v1,
 	struct ent *v2);
-extern	void copyent(register struct ent *n, register struct ent *p, int dr,
-	int dc);
+extern	void copyent(register struct ent *n, register struct ent *p,
+	int dr, int dc, int r1, int c1, int r2, int c2, int transpose);
 extern	void creadfile(char *save, int  eraseflg);
 extern	void decompile(register struct enode *e, int priority);
 extern	void deleterow(register int arg);
@@ -433,6 +438,7 @@ extern	void fix_frames(int row1, int col1, int row2, int col2,
 extern	void fix_ranges(int row1, int col1, int row2, int col2,
 	int delta1, int delta2);
 extern	void flush_saved();
+extern	void formatcol(int arg);
 extern	void format_cell(struct ent *v1, struct ent *v2, char *s);
 extern	void forwcol(int arg);
 extern	void forwrow(int arg);
@@ -474,6 +480,7 @@ extern	void printfile(char *fname, int r0, int c0, int rn, int cn);
 extern	void pullcells(int to_insert);
 extern	void query(char *s, char *data);
 extern	void readfile(char *fname, int eraseflg);
+extern	void remember(int save);
 extern	void resetkbd();
 extern	void rowshow_op();
 extern	void scxfree(char *p);
@@ -524,6 +531,7 @@ extern	int Crypt;
 #endif
 extern	char *mdir;
 extern	char *autorun;
+extern	int skipautorun;
 extern	char *fkey[FKEYS];
 extern	char *scext;
 extern	char *ascext;
@@ -548,6 +556,8 @@ extern	int showtop;
 extern	int color;
 extern	int colorneg;
 extern	int colorerr;
+extern	int braille;
+extern	int braillealt;
 extern	int loading;
 extern	int getrcqual;
 extern	int tbl_style;
