@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.13 $
+ *		$Revision: 7.15 $
  *
  */
 
@@ -89,6 +89,7 @@ update(int anychanged)		/* did any cell really change in value? */
     struct crange		*cr;
     int				ftoprows, fbottomrows, fleftcols, frightcols;
     int				ftrows, fbrows, flcols, frcols;
+    bool			message;
 
 #ifndef MSDOS
     /*
@@ -608,15 +609,17 @@ update(int anychanged)		/* did any cell really change in value? */
 		continue;
 	    (void) move(2, col);
 	    k = (fwidth[i] - strlen(coltoa(i)))/2;
-	    if (k == 0)
-		(void) printw("%1s", coltoa(i));
+	    if (fwidth[i] == 1)
+		(void) printw("%1s", coltoa(i%26));
 	    else
-	        (void) printw("%*s%-*s", k, " ", fwidth[i]-k, coltoa(i));
+	        (void) printw("%*s%-*s", k, "", fwidth[i]-k, coltoa(i));
 	    col += fwidth[i];
 	}
 	(void) standend();
     }
 
+    (void) move(1, 0);
+    message = (inch() & A_CHARTEXT) != ' ';
     if (showrange) {
 	if (showrange == SHOWROWS) {
 	    minsr = showsr < currow ? showsr : currow;
@@ -624,8 +627,7 @@ update(int anychanged)		/* did any cell really change in value? */
 	    maxsr = showsr > currow ? showsr : currow;
 	    maxsc = fr ? fr->or_right->col : maxcols;
 
-	    if (showtop) {
-		(void) move(1,0);
+	    if (showtop && !message) {
 		(void) clrtoeol();
 		(void) printw("Default range:  %d:%d", minsr, maxsr);
 	    }
@@ -635,13 +637,12 @@ update(int anychanged)		/* did any cell really change in value? */
 	    maxsr = maxrows;
 	    maxsc = showsc > curcol ? showsc : curcol;
 
-	    if (showtop) {
+	    if (showtop && !message) {
 		char r[6];
 
 		strcpy(r, coltoa(minsc));
 		strcat(r, ":");
 		strcat(r, coltoa(maxsc));
-		(void) move(1,0);
 		(void) clrtoeol();
 		(void) printw("Default range:  %s", r);
 	    }
@@ -651,16 +652,17 @@ update(int anychanged)		/* did any cell really change in value? */
 	    maxsr = showsr > currow ? showsr : currow;
 	    maxsc = showsc > curcol ? showsc : curcol;
 
-	    if (showtop) {
-		(void) move(1,0);
+	    if (showtop && !message) {
 		(void) clrtoeol();
 		(void) printw("Default range:  %s",
 			    r_name(minsr, minsc, maxsr, maxsc));
 	    }
 	}
+    } else if (braille && braillealt && !message && mode_ind == 'v') {
+	(void) clrtoeol();
+	(void) printw("Current cell:   %s%d ", coltoa(curcol), currow);
     }
 
-
     /* Repaint the visible screen */
     if (showrange || anychanged || FullUpdate || standlast) {
 	/* may be reset in loop, if not next time we will do a FullUpdate */
@@ -866,8 +868,8 @@ update(int anychanged)		/* did any cell really change in value? */
 			(void) printw("%*s", fwidth[col], " ");
 		    }
 		} /* else */
-	    }
-	    else if (!*pp && color && has_colors && cr && cr->r_color != 1) {
+	    } else
+	    if (!*pp && color && has_colors && cr && cr->r_color != 1) {
 		move(r, c);
 		color_set(cr->r_color, NULL);
 		printw("%*s", fwidth[col], " ");
@@ -927,7 +929,14 @@ update(int anychanged)		/* did any cell really change in value? */
 	(void) addch(mode_ind);
 	(void) addstr("> ");
 	(void) addstr(line);
-	(void) move((linelim+3+ctlchars)/cols, (linelim+3+ctlchars)%cols);
+	if (!braille || (!message && mode_ind != 'v'))
+	    (void) move((linelim+3+ctlchars)/cols, (linelim+3+ctlchars)%cols);
+	else if (message)
+	    move(1, 0);
+	else if (braillealt)
+	    move(1, 16);
+	else
+	    move(lastmy,lastmx+(fwidth[lastcol]-strlen(coltoa(lastcol)))/2);
     } else {
 	if (showtop) {			/* show top line */
 	    register struct ent *p1;
@@ -1000,7 +1009,14 @@ update(int anychanged)		/* did any cell really change in value? */
 	    if (p1 && p1->flags&is_locked)
 		(void) addstr(" locked");
 	}
-	if (showcell)
+	if (braille)
+	    if (message)
+		move(1, 0);
+	    else if (braillealt)
+		move(0, 0);
+	    else
+		move(lastmy,lastmx+(fwidth[lastcol]-strlen(coltoa(lastcol)))/2);
+	else if (showcell)
 	    move(lines - 1, cols - 1);
 	else
 	    (void) move(lastmy, lastmx+fwidth[lastcol]);
@@ -1014,7 +1030,12 @@ update(int anychanged)		/* did any cell really change in value? */
 	(void) clrtoeol();	/* get rid of topline display */
 	(void) printw(revmsg);
 	*revmsg = '\0';		/* don't show it again */
-	if (showcell)
+	if (braille)
+	    if (message)
+		move(1, 0);
+	    else
+		move(lastmy,lastmx+(fwidth[lastcol]-strlen(coltoa(lastcol)))/2);
+	else if (showcell)
 	    move(lines - 1, cols - 1);
 	else
 	    (void) move(lastmy, lastmx + fwidth[lastcol]);
