@@ -1,7 +1,7 @@
 /*	SC	A Spreadsheet Calculator
  *
  *	One line vi emulation
- *	$Revision: 7.13 $
+ *	$Revision: 7.15 $
  */
 
 #include <sys/types.h>
@@ -37,6 +37,7 @@ static void for_hist();
 static int  for_line(int arg, int stop_null);
 static int  for_word(int arg, int end_word, int stop_null);
 static void last_col();
+static void match_paren();
 static void change_case(int arg);
 static void rep_char();
 static void replace_in_line(int c);
@@ -84,6 +85,8 @@ static int	undo_lim;
 static char	dotb[DOTLEN];
 static int	doti = 0;
 static int	do_dot = 0;
+static int	nosavedot = 1;
+static int	dotarg = 1;
 static char	putbuf[FBUFLEN];
 static int	findfunc = '0';
 static int	findchar = 1;
@@ -95,8 +98,10 @@ write_line(int c)
     struct frange *fr;
     struct crange *cr;
 
+    error("");
     if (c != ctl('i')) completethis = NULL;
     if (mode == EDIT_MODE) {
+	nosavedot = 0;
 	switch (c) {
 	case KEY_BACKSPACE:
 	case (ctl('h')):	linelim = back_line(arg);		break;
@@ -132,13 +137,12 @@ write_line(int c)
 			    linelim = to_char(arg, -(finddir));
 									break;
 	case '~':	u_save(c); change_case(arg);			break;
+	case '%':	match_paren();					break;
 #ifdef KEY_FIND
 	case KEY_FIND:
 #endif
 	case '/':	search_mode();					break;
-#ifdef KEY_HOME
 	case KEY_HOME:
-#endif
 	case (ctl('a')):
 	case '0':	col_0();					break;
 	case 'A':	u_save(c); last_col(); append_line();		break;
@@ -201,13 +205,45 @@ write_line(int c)
 	case (ctl('m')):	cr_line();				break;
 	case (ctl('v')):	toggle_navigate_mode();			break;
 	case KEY_LEFT:
-	case (ctl('b')):	linelim = back_line(arg);		break;
+	case (ctl('b')):	if (numeric_field) {
+				    c = craction;
+				    craction = 0;
+				    cr_line();
+				    numeric_field = 0;
+				    craction = c;
+				    backcol(1);
+				} else
+				linelim = back_line(arg);		break;
 	case KEY_RIGHT:
-	case (ctl('f')):	linelim = for_line(arg, 1);		break;
+	case (ctl('f')):	if (numeric_field) {
+				    c = craction;
+				    craction = 0;
+				    cr_line();
+				    numeric_field = 0;
+				    craction = c;
+				    forwcol(1);
+				} else
+				linelim = for_line(arg, 1);		break;
 	case KEY_DOWN:
-	case (ctl('n')):	for_hist();				break;
+	case (ctl('n')):	if (numeric_field) {
+				    c = craction;
+				    craction = 0;
+				    cr_line();
+				    numeric_field = 0;
+				    craction = c;
+				    forwrow(1);
+				} else
+				for_hist();				break;
 	case KEY_UP:
-	case (ctl('p')):	back_hist();				break;
+	case (ctl('p')):	if (numeric_field) {
+				    c = craction;
+				    craction = 0;
+				    cr_line();
+				    numeric_field = 0;
+				    craction = c;
+				    backrow(1);
+				} else
+				back_hist();				break;
 	case KEY_HOME:
 	case (ctl('a')):	col_0();				break;
 	case KEY_END:
@@ -260,7 +296,8 @@ write_line(int c)
 				    toggle_navigate_mode();
 				    cr_line();
 				}					break;
-	case (ctl('a')):	{   /* insert variable value */
+	case 'v':
+	case (ctl('v')):	{   /* insert variable value */
 				    struct ent *p = *ATBL(tbl, currow, curcol);
 				    char temp[100];
 
@@ -268,11 +305,9 @@ write_line(int c)
 					(void) sprintf(temp, "%.*f",
 						precision[curcol], p->v);
 					ins_string(temp);
+					toggle_navigate_mode();
 				    }
 				}					break;
-	case 'v':
-	case (ctl('v')):	ins_string(v_name(currow, curcol));
-				toggle_navigate_mode();			break;
 	case 'c':		if ((cr = find_crange(currow, curcol))) {
 				    ins_string(r_name(cr->r_left->row,
 					    cr->r_left->col,
@@ -347,71 +382,50 @@ write_line(int c)
 	case 'L':
 				forwcol(lcols -(curcol-stcol)+1);
 									break;
+	case (ctl('a')):
 	case KEY_HOME:
-				/* Remember the current position */
-				savedrow[0] = currow;
-				savedcol[0] = curcol;
-				savedstrow[0] = strow;
-				savedstcol[0] = stcol;
-
+				remember(0);
 				currow = 0;
 				curcol = 0;
 				rowsinrange = 1;
 				colsinrange = fwidth[curcol];
-				FullUpdate++;
-									break;
+				remember(1);
+				FullUpdate++;				break;
 	case '0':
-				/* Remember the current position */
-				savedrow[0] = currow;
-				savedcol[0] = curcol;
-				savedstrow[0] = strow;
-				savedstcol[0] = stcol;
-
+				remember(0);
 				curcol = 0;
 		   		rowsinrange = 1;
-		   		colsinrange = fwidth[curcol];		break;
+		   		colsinrange = fwidth[curcol];
+				remember(1);				break;
 	case '$':		{
 				register struct ent *p;
 
-				/* Remember the current position */
-				savedrow[0] = currow;
-				savedcol[0] = curcol;
-				savedstrow[0] = strow;
-				savedstcol[0] = stcol;
-
+				remember(0);
 				curcol = maxcols - 1;
 				while (!VALID_CELL(p, currow, curcol) &&
 					curcol > 0)
 				    curcol--;
 				rowsinrange = 1;
 				colsinrange = fwidth[curcol];
-				break;
+				remember(1);				break;
 				}
 	case '^':
-				/* Remember the current position */
-				savedrow[0] = currow;
-				savedcol[0] = curcol;
-				savedstrow[0] = strow;
-				savedstcol[0] = stcol;
-
+				remember(0);
 				currow = 0;
 				rowsinrange = 1;
-				colsinrange = fwidth[curcol];		break;
+				colsinrange = fwidth[curcol];
+				remember(1);				break;
 	case '#':		{
 				register struct ent *p;
 
-				/* Remember the current position */
-				savedrow[0] = currow;
-				savedcol[0] = curcol;
-				savedstrow[0] = strow;
-				savedstcol[0] = stcol;
-
+				remember(0);
 				currow = maxrows - 1;
 				while (!VALID_CELL(p, currow, curcol) &&
 					currow > 0)
 				    currow--;
 				rowsinrange = 1;
-				colsinrange = fwidth[curcol];		break;
+				colsinrange = fwidth[curcol];
+				remember(1);				break;
 				}
 	case 'm':
 				markcell();				break;
@@ -470,6 +484,10 @@ write_line(int c)
 				colsinrange = fwidth[curcol];
 				break;
 				}
+	case 'C':		if (braille) {
+				    braillealt ^= 1;
+				    break;
+				}
 	}
     }
 }
@@ -481,6 +499,7 @@ edit_mode()
     mode_ind = 'e';
     if (linelim < 0)	/* -1 says stop editing, ...so we still aren't */
 	return;
+    numeric_field = 0;
     linelim = back_line(1);
 }
 
@@ -550,14 +569,14 @@ dotab()
     static struct range *nextmatch;
     int len;
 
-    if (linelim > 1 && isalnum(line[linelim-1]) || line[linelim-1] == '_' ||
+    if (linelim > 0 && isalnum(line[linelim-1]) || line[linelim-1] == '_' ||
 	    (completethis && line[linelim-1] == ' ')) {
 	if (!completethis) {
 	    for (completethis = line + linelim - 1; isalnum(*completethis) ||
 		    *completethis == '_'; completethis--) /* */;
 	    completethis++;
 	    len = line + linelim - completethis;
-	    if (!find_range(completethis, len, NULL, NULL, &lastmatch)) {
+	    if (!find_range(completethis, -len, NULL, NULL, &lastmatch)) {
 		firstmatch = lastmatch;
 		while (firstmatch->r_next &&
 			!strncmp(completethis, firstmatch->r_next->r_name, len))
@@ -643,9 +662,10 @@ showdr()
 static void
 savedot(int c)
 {
-    if (do_dot || (c == '\n'))
+    if (do_dot || nosavedot || (c == '\n'))
 	return;
 
+    if (doti == 0) dotarg = arg;
     if (doti < DOTLEN-1) {
 	if (c > 255) {
 	    if (doti < DOTLEN-2) {
@@ -670,6 +690,10 @@ dotcmd()
 	return;
     do_dot = 1;
     doti = 0;
+    if (arg == 1)
+	arg = dotarg;
+    else
+	dotarg = arg;
     while (dotb[doti] != '\0') {
 	if ((c = dotb[doti++]) < 4)
 	    c = c * 256 + dotb[doti++];
@@ -695,6 +719,7 @@ vigetch()
 	    return (nmgetch());
 	}
     }
+    update(1);
     c = nmgetch();
     return (c);
 }
@@ -1013,16 +1038,21 @@ get_motion(int change)
     int arg2 = 0;
 
     c = vigetch();
-    savedot(c);
-    if (c == 0)		return (0);
+    if (c == 0) {
+	savedot(c);
+	return (0);
+    }
     while (c >= '0' && c <= '9') {
 	arg2 = 10 * arg2 + c - '0';
 	c = vigetch();
-	savedot(c);
     }
     if (!arg2)
 	arg2++;
     arg *= arg2;
+    if (!nosavedot) {
+	savedot(c);
+	dotarg = arg;
+    }
     switch (c) {
 	case '$':	return (strlen(line));
 	case 'b':	return (back_word(arg));
@@ -1088,10 +1118,11 @@ cr_line()
 
     insert_mode();
     if (linelim != -1) {
-	showrange = 0;
 	save_hist();
+	nosavedot = 1;
 	linelim = 0;
 	(void) yyparse();
+	showrange = 0;
 	linelim = -1;
 	if (cellassign) {
 	    cellassign = 0;
@@ -1142,20 +1173,22 @@ cr_line()
 static void
 save_hist()
 {
-    if (lasthist < 0)
-	lasthist = 1;
-    else
-	lasthist = lasthist % HISTLEN + 1;
+    if (!lasthist || strcmp(history[lasthist].histline, line)) {
+	if (lasthist < 0)
+	    lasthist = 1;
+	else
+	    lasthist = lasthist % HISTLEN + 1;
 
-    if (lasthist > endhist)
-	endhist = lasthist;
+	if (lasthist > endhist)
+	    endhist = lasthist;
 
-    if (history[lasthist].len < strlen(line)+1) {
-    	history[lasthist].len = strlen(line)+40;
-	history[lasthist].histline = scxrealloc(history[lasthist].histline,
-		history[lasthist].len);
+	if (history[lasthist].len < strlen(line)+1) {
+	    history[lasthist].len = strlen(line)+40;
+	    history[lasthist].histline = scxrealloc(history[lasthist].histline,
+		    history[lasthist].len);
+	}
+	(void) strcpy(history[lasthist].histline, line);
     }
-    (void) strcpy(history[lasthist].histline, line);
     if (history[0].histline) {
 	scxfree(history[0].histline);
 	history[0].histline = (void *)0;
@@ -1317,6 +1350,58 @@ to_char(int arg, int n)
     return (i);
 }
 
+static void
+match_paren()
+{
+    register int i;
+    int nest = 1;
+    int tmp = linelim;
+
+    if (line[linelim] == '(') {
+	while (nest && ++linelim >= 0 && line[linelim]) {
+	    if (line[linelim] == '(')
+		nest++;
+	    else if (line[linelim] == ')')
+		nest--;
+	}
+	if (line[linelim] != ')')
+	    linelim = tmp;
+    }
+    else if (line[linelim] == ')') {
+	while (nest && --linelim >= 0 && line[linelim]) {
+	    if (line[linelim] == ')')
+		nest++;
+	    else if (line[linelim] == '(')
+		nest--;
+	}
+	if (line[linelim] != '(')
+	    linelim = tmp;
+    }
+}
+
+/* If save is 0, remember the current position.  Otherwise, if the current
+ * cell has changed since the last remember(0), save the remembered location
+ * for the `, ', and c comands.
+ */
+void
+remember(int save)
+{
+    static int remrow, remcol, remstrow, remstcol;
+
+    if (save && (currow != remrow || curcol != remcol ||
+	    strow != remstrow || stcol != remstcol)) {
+	savedrow[0] = remrow;
+	savedcol[0] = remcol;
+	savedstrow[0] = remstrow;
+	savedstcol[0] = remstcol;
+    } else {
+	remrow = currow;
+	remcol = curcol;
+	remstrow = strow;
+	remstcol = stcol;
+    }
+}
+
 static void
 dogoto()
 {
@@ -1390,7 +1475,6 @@ query(char *s, char *data)
 		FullUpdate++;
 		clearok(stdscr,1);
 		update(1);
-		if (s != NULL) error(s);
 		break;
 	    default:
 		write_line(c);
